# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Optional, Tuple, Union, cast

import torch

from theseus.core import CostFunction, CostWeight, Variable
from theseus.geometry import SE2, OptionalJacobians, Point2
from theseus.geometry.so2 import SO2


# Based on the model described in
# Zhou et al. A Fast Stochastic Contact Model for Planar Pushing and Grasping:
# Theory and Experimental Validation, 2017
# https://arxiv.org/abs/1705.10664
class QuasiStaticPushingPlanar(CostFunction):
    def __init__(
        self,
        obj1: SE2,
        obj2: SE2,
        eff1: SE2,
        eff2: SE2,
        c_square: Union[Variable, torch.Tensor, float],
        cost_weight: CostWeight,
        name: Optional[str] = None,
    ):
        super().__init__(cost_weight, name=name)
        self.obj1 = obj1
        self.obj2 = obj2
        self.eff1 = eff1
        self.eff2 = eff2
        self.register_optim_vars(["obj1", "obj2", "eff1", "eff2"])

        if not isinstance(c_square, Variable):
            if isinstance(c_square, float):
                c_square = torch.tensor(c_square, dtype=self.obj1.dtype).view(1, 1)
            c_square = Variable(c_square, name=f"csquare_{name}")
        if c_square.tensor.squeeze().ndim > 1:
            raise ValueError("dt must be a 0-D or 1-D tensor.")
        self.c_square = c_square
        self.register_aux_vars(["c_square"])

    def _get_tensor_info(self) -> Tuple[int, torch.device, torch.dtype]:
        return self.obj1.shape[0], self.obj1.device, self.obj1.dtype

    # See Eqs. 3-7 in Zhou et al.
    def _compute_D(
        self,
        contact_point_2: torch.Tensor,
        J_cp2_e2: OptionalJacobians,
        get_jacobians: bool = False,
    ) -> Tuple[torch.Tensor, OptionalJacobians]:
        batch_size, device, dtype = self._get_tensor_info()

        # Setting up intermediate jacobians, if needed
        J_cp2o: OptionalJacobians = [] if get_jacobians else None

        # Current contact point in object frame
        contact_point_2__obj = self.obj2.transform_to(contact_point_2, jacobians=J_cp2o)
        px = contact_point_2__obj.tensor[:, 0]
        py = contact_point_2__obj.tensor[:, 1]

        # Putting D together
        D = (
            torch.eye(3, 3, device=device, dtype=dtype)
            .unsqueeze(0)
            .repeat(batch_size, 1, 1)
        )
        D[:, 0, 2] = D[:, 2, 0] = -py
        D[:, 1, 2] = D[:, 2, 1] = px
        D[:, 2, 2] = -self.c_square.tensor.view(-1)

        if not get_jacobians:
            return D, None

        # Computing jacobians
        J_cp2o_o2, J_cp2o_cp2 = J_cp2o
        J_cp2o_e2 = J_cp2o_cp2.matmul(J_cp2_e2[0])

        def _get_J_D_var(J_cp2o_var_):
            # For matmul batching, this is set up with dimensions:
            #   batch_size x se2_dim x 3 x 3,
            # so that J_D_var_[b, d] is the (3, 3) matrix derivative
            # of the b-th batch D wrt to the d-th dimention of var
            J_D_var_ = torch.zeros(batch_size, 3, 3, 3, device=device, dtype=dtype)
            J_D_var_[..., 0, 2] = J_D_var_[..., 2, 0] = -J_cp2o_var_[:, 1]
            J_D_var_[..., 1, 2] = J_D_var_[..., 2, 1] = J_cp2o_var_[:, 0]
            return J_D_var_

        J_D_o2 = _get_J_D_var(J_cp2o_o2)
        J_D_e2 = _get_J_D_var(J_cp2o_e2)

        return D, [J_D_o2, J_D_e2]

    # See Eqs. 3-7 in Zhou et al.
    def _compute_V(
        self,
        obj2_angle: SO2,
        J_o2angle_o2: OptionalJacobians,
        get_jacobians: bool = False,
    ) -> Tuple[torch.Tensor, OptionalJacobians]:
        batch_size, device, dtype = self._get_tensor_info()

        # Setting up intermediate jacobians, if needed
        J_o2xy_o2: OptionalJacobians = [] if get_jacobians else None
        J_o1xy_o1: OptionalJacobians = [] if get_jacobians else None
        J_vxyoo: OptionalJacobians = [] if get_jacobians else None
        J_odw: OptionalJacobians = [] if get_jacobians else None
        J_omega_odw: OptionalJacobians = [] if get_jacobians else None

        # Compute object velocities using consecutive object poses
        vel_xy_obj__world = self.obj2.xy(jacobians=J_o2xy_o2) - self.obj1.xy(
            jacobians=J_o1xy_o1
        )
        vel_xy_obj__obj = obj2_angle.unrotate(vel_xy_obj__world, jacobians=J_vxyoo)

        # Putting V together
        obj_diff__world = cast(SE2, self.obj1.between(self.obj2, jacobians=J_odw))
        vx = vel_xy_obj__obj.tensor[:, 0]
        vy = vel_xy_obj__obj.tensor[:, 1]
        omega = obj_diff__world.theta(jacobians=J_omega_odw)
        V = torch.stack([vx, vy, omega.squeeze(1)], dim=1)

        if not get_jacobians:
            return V, None

        # Computing jacobians
        def _get_J_V_var(J_vxyoo_var_, J_omega_var_):
            # For matmul batching, this is set up with dimensions:
            #   batch_size x se2_dim x 3 x 1,
            # so that J_V_var_[b, d] is the (3, 1) vector derivative
            # of the b-th batch V wrt to the d-th dimention of var
            J_V_var_ = torch.zeros(batch_size, 3, 3, device=device, dtype=dtype)
            J_V_var_[..., 0] = J_vxyoo_var_[:, 0]
            J_V_var_[..., 1] = J_vxyoo_var_[:, 1]
            J_V_var_[..., 2] = J_omega_var_[:, 0]
            return J_V_var_.unsqueeze(3)

        # Computing velocity derivates wrt to objects 1, 2 variables
        J_vxyow_o2 = J_o2xy_o2[0]
        J_vxyow_o1 = -J_o1xy_o1[0]
        J_vxyoo_o2angle, J_vxyoo_vxyow = J_vxyoo
        J_vxyoo_o1 = J_vxyoo_vxyow.matmul(J_vxyow_o1)
        J_vxyoo_o2 = J_vxyoo_o2angle.matmul(J_o2angle_o2[0]) + J_vxyoo_vxyow.matmul(
            J_vxyow_o2
        )

        J_odw_o1, J_odw_o2 = J_odw
        J_omega_o1 = J_omega_odw[0].matmul(J_odw_o1)
        J_omega_o2 = J_omega_odw[0].matmul(J_odw_o2)

        J_V_o1 = _get_J_V_var(J_vxyoo_o1, J_omega_o1)
        J_V_o2 = _get_J_V_var(J_vxyoo_o2, J_omega_o2)

        return V, [J_V_o1, J_V_o2]

    # See Eqs. 3-7 in Zhou et al.
    def _compute_Vp(
        self,
        obj2_angle: SO2,
        J_o2angle_o2: OptionalJacobians,
        contact_point_2: torch.Tensor,
        J_cp2_e2: OptionalJacobians,
        get_jacobians: bool = False,
    ) -> Tuple[torch.Tensor, List[torch.Tensor]]:
        batch_size, device, dtype = self._get_tensor_info()

        # Setting up intermediate jacobians, if needed
        J_cp1_e1: OptionalJacobians = [] if get_jacobians else None
        J_vxyco: OptionalJacobians = [] if get_jacobians else None

        # Compute contact point velocities using consecutive end effector poses
        contact_point_1 = self.eff1.xy(jacobians=J_cp1_e1)
        vel_xy_contact__world = cast(Point2, contact_point_2 - contact_point_1)

        # Transform velocity to object's 2 frame
        vel_xy_contact__obj = obj2_angle.unrotate(
            vel_xy_contact__world, jacobians=J_vxyco
        )

        # Putting Vp together
        v_px = vel_xy_contact__obj.tensor[:, 0]
        v_py = vel_xy_contact__obj.tensor[:, 1]
        Vp = torch.stack([v_px, v_py, torch.zeros_like(v_px)], dim=1)

        if not get_jacobians:
            return Vp, None

        # Computing jacobians
        def _get_J_Vp_var(J_vxyco_var_):
            # For matmul batching, this is set up with dimensions:
            #   batch_size x se2_dim x 3 x 1,
            # so that J_Vp_var_[b, d] is the (3, 1) vector derivative
            # of the b-th batch Vp wrt to the d-th dimention of var
            J_Vp_var_ = torch.zeros(batch_size, 3, 3, device=device, dtype=dtype)
            J_Vp_var_[..., 0] = J_vxyco_var_[:, 0]  # batch_size x se2_dim
            J_Vp_var_[..., 1] = J_vxyco_var_[:, 1]  # batch_size x se2_dim
            return J_Vp_var_.unsqueeze(3)

        J_vxyco_o2angle, J_vxyco_vxycw = J_vxyco
        J_vxyco_o2 = J_vxyco_o2angle.matmul(J_o2angle_o2[0])
        J_vxyco_e1 = J_vxyco_vxycw.matmul(-J_cp1_e1[0])
        J_vxyco_e2 = J_vxyco_vxycw.matmul(J_cp2_e2[0])

        J_Vp_o2 = _get_J_Vp_var(J_vxyco_o2)
        J_Vp_e1 = _get_J_Vp_var(J_vxyco_e1)
        J_Vp_e2 = _get_J_Vp_var(J_vxyco_e2)

        return Vp, [J_Vp_o2, J_Vp_e1, J_Vp_e2]

    def _error_and_jacobians_impl(
        self, get_jacobians: bool
    ) -> Tuple[torch.Tensor, OptionalJacobians]:
        J_o2angle_o2: OptionalJacobians = [] if get_jacobians else None
        J_cp2_e2: OptionalJacobians = [] if get_jacobians else None

        # These quantities are needed by two or more of D, V, Vp, so we compute once.
        obj2_angle = self.obj2.rotation
        self.obj2.theta(jacobians=J_o2angle_o2)
        contact_point_2 = self.eff2.xy(jacobians=J_cp2_e2)

        # D * V = Vp, See Zhou et al.
        # A Fast Stochastic Contact Model for Planar Pushing and Grasping:
        # Theory and Experimental Validation, 2017
        # https://arxiv.org/abs/1705.10664
        D, J_D = self._compute_D(contact_point_2, J_cp2_e2, get_jacobians=get_jacobians)
        V, J_V = self._compute_V(obj2_angle, J_o2angle_o2, get_jacobians=get_jacobians)
        Vp, J_Vp = self._compute_Vp(
            obj2_angle,
            J_o2angle_o2,
            contact_point_2,
            J_cp2_e2,
            get_jacobians=get_jacobians,
        )

        error = torch.bmm(D, V.unsqueeze(2)).squeeze(2) - Vp

        if not get_jacobians:
            return error, None

        # Computing jacobians
        V_exp = V.view(-1, 1, 3, 1)
        D_exp = D.view(-1, 1, 3, 3)
        sum_str = "bdij,bdjk->bdik"

        J_D_o2, J_D_e2 = J_D
        J_V_o1, J_V_o2 = J_V
        J_Vp_o2, J_Vp_e1, J_Vp_e2 = J_Vp

        def _get_J_err_var_(J_D_var_, J_V_var_, J_Vp_var_):
            # After einsum, each result is a tensor with size:
            #   batch_size x se2_dim x 3 x 1
            # so that after all terms are summed,
            # J_err_var_[b, d] is a (3, 1) vector derivative of
            # the b-th batch element of the 3-D error wrt to the
            # d-th dimension of the optimization variable.
            # At return, we squeeze the trailing dimension, and permute, so that the
            # output has shape (batch_size, 3, se2_dim), which is the expected jacobian
            # shape.
            J_err_var_ = torch.zeros_like(J_V_o1)
            if J_D_var_ is not None:
                J_err_var_ += torch.einsum(sum_str, J_D_var_, V_exp)
            if J_V_var_ is not None:
                J_err_var_ += torch.einsum(sum_str, D_exp, J_V_var_)
            if J_Vp_var_ is not None:
                J_err_var_ += -J_Vp_var_
            return J_err_var_.squeeze(3).permute(0, 2, 1)

        J_err_o1 = _get_J_err_var_(None, J_V_o1, None)
        J_err_o2 = _get_J_err_var_(J_D_o2, J_V_o2, J_Vp_o2)
        J_err_e1 = _get_J_err_var_(None, None, J_Vp_e1)
        J_err_e2 = _get_J_err_var_(J_D_e2, None, J_Vp_e2)

        return error, [J_err_o1, J_err_o2, J_err_e1, J_err_e2]

    def error(self) -> torch.Tensor:
        return self._error_and_jacobians_impl(get_jacobians=False)[0]

    def jacobians(self) -> Tuple[List[torch.Tensor], torch.Tensor]:
        error, jacobians = self._error_and_jacobians_impl(get_jacobians=True)
        return jacobians, error

    def dim(self) -> int:
        return 3

    def _copy_impl(self, new_name: Optional[str] = None) -> "QuasiStaticPushingPlanar":
        return QuasiStaticPushingPlanar(
            self.obj1.copy(),
            self.obj2.copy(),
            self.eff1.copy(),
            self.eff2.copy(),
            self.c_square.copy(),
            self.weight.copy(),
            name=new_name,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import abc
from typing import Dict, Optional, Union

import differentiable_robot_model as drm
import torch

from theseus.geometry import SE3, LieGroup, Point2, Vector

RobotModelInput = Union[torch.Tensor, Vector]


class KinematicsModel(abc.ABC):
    def __init__(self):
        pass

    @abc.abstractmethod
    def forward_kinematics(self, robot_pose: RobotModelInput) -> Dict[str, LieGroup]:
        pass


class IdentityModel(KinematicsModel):
    def __init__(self):
        super().__init__()

    def forward_kinematics(self, robot_pose: RobotModelInput) -> Dict[str, LieGroup]:
        if isinstance(robot_pose, Point2) or isinstance(robot_pose, Vector):
            assert robot_pose.dof() == 2
            return {"state": robot_pose}
        raise NotImplementedError(
            f"IdentityModel not implemented for pose with type {type(robot_pose)}."
        )


class UrdfRobotModel(KinematicsModel):
    def __init__(self, urdf_path: str):
        self.drm_model = drm.DifferentiableRobotModel(urdf_path)

    def _postprocess_quaternion(self, quat):
        # Convert quaternion convention (DRM uses xyzw, Theseus uses wxyz)
        quat_converted = torch.cat([quat[..., 3:], quat[..., :3]], dim=-1)

        # Normalize quaternions
        quat_normalized = quat_converted / torch.linalg.norm(
            quat_converted, dim=-1, keepdim=True
        )

        return quat_normalized

    def forward_kinematics(
        self,
        joint_states: RobotModelInput,
        jacobians: Optional[Dict[str, Optional[torch.Tensor]]] = None,
    ) -> Dict[str, LieGroup]:
        """Computes forward kinematics
        Args:
            joint_states: Vector of all joint angles
        Outputs:
            Dictionary that maps link name to link pose
        """
        # Check input dimensions
        robot_model_dofs = len(self.drm_model.get_joint_limits())
        assert joint_states.shape[-1] == robot_model_dofs, (
            f"Robot model dofs ({robot_model_dofs}) incompatible with "
            f"input joint state dimensions ({joint_states.shape[-1]})."
        )

        # Parse input
        if isinstance(joint_states, torch.Tensor):
            joint_states_input = joint_states
        elif isinstance(joint_states, Vector):
            joint_states_input = joint_states.tensor
        else:
            raise Exception(
                "Invalid input joint states data type. "
                "Valid types are torch.Tensor and th.Vector."
            )

        # Compute forward kinematics for all links
        fk_output = self.drm_model.compute_forward_kinematics_all_links(
            joint_states_input
        )

        link_poses: Dict[str, LieGroup] = {}
        for link_name in self.drm_model.get_link_names():
            pos, quat = fk_output[link_name]
            quat_processed = self._postprocess_quaternion(quat)

            link_poses[link_name] = SE3(
                x_y_z_quaternion=torch.cat([pos, quat_processed], dim=-1)
            )

        # Compute jacobians
        if jacobians is not None:
            for link_name in jacobians.keys():
                jac_lin, jac_rot = self.drm_model.compute_endeffector_jacobian(
                    joint_states_input, link_name
                )
                jacobians[link_name] = torch.cat([jac_lin, jac_rot], dim=-2)

        return link_poses



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
from typing import Dict, List, Optional, Tuple

import torch

import theseus as th


class MotionPlanner:
    def __init__(
        self,
        map_size: int,
        epsilon_dist: float,
        total_time: float,
        collision_weight: float,
        Qc_inv: List[List[int]],
        num_time_steps: int,
        optim_method: str,
        max_optim_iters: int,
        step_size: float = 1.0,
        use_single_collision_weight: bool = True,
        device: str = "cpu",
        dtype: torch.dtype = torch.double,
    ):
        self.map_size = map_size
        self.epsilon_dist = epsilon_dist
        self.total_time = total_time
        self.collision_weight = collision_weight
        self.Qc_inv = copy.deepcopy(Qc_inv)
        self.num_time_steps = num_time_steps
        self.optim_method = optim_method
        self.max_optim_iters = max_optim_iters
        self.step_size = step_size
        self.use_single_collision_weight = use_single_collision_weight
        self.device = device
        self.dtype = dtype

        self.trajectory_len = num_time_steps + 1

        # --------------------------------------------------------------------------- #
        # ---------------------------- Auxiliary variables -------------------------- #
        # --------------------------------------------------------------------------- #
        # First we create auxiliary variables for all the cost functions we are going
        # to need. Auxiliary variables are variables whose value won't be changed by
        # the motion planner's optimizer, but that are required to compute cost
        # functions.
        # By giving them names, we can update for each batch (if needed),
        # via the motion planner's forward method.
        sdf_origin = th.Point2(name="sdf_origin")
        start_point = th.Point2(name="start")
        goal_point = th.Point2(name="goal")
        cell_size_tensor = th.Variable(torch.empty(1, 1), name="cell_size")
        sdf_data_tensor = th.Variable(
            torch.empty(1, map_size, map_size), name="sdf_data"
        )
        cost_eps = th.Variable(torch.tensor(epsilon_dist).view(1, 1), name="cost_eps")
        dt = th.Variable(
            torch.tensor(total_time / num_time_steps).view(1, 1), name="dt"
        )

        # --------------------------------------------------------------------------- #
        # ------------------------------- Cost weights ------------------------------ #
        # --------------------------------------------------------------------------- #
        # For the GP cost functions, we create a single GPCost weight
        gp_cost_weight = th.eb.GPCostWeight(torch.tensor(Qc_inv), dt)

        # Now we create cost weights for the collision-avoidance cost functions
        # Each of this is a scalar cost weight with a named auxiliary variable.
        # Before running the motion planner, each cost weight value can be updated
        # by passing {name: new_cost_weight_tensor} to the forward method
        collision_cost_weights = []
        if use_single_collision_weight:
            collision_cost_weights.append(
                th.ScaleCostWeight(
                    th.Variable(torch.tensor(collision_weight), name="collision_w")
                )
            )
        else:
            for i in range(1, self.trajectory_len):
                collision_cost_weights.append(
                    th.ScaleCostWeight(
                        th.Variable(
                            torch.tensor(collision_weight), name=f"collision_w_{i}"
                        )
                    )
                )

        # For hard-constraints (end points pos/vel) we use a single scalar weight
        # with high value
        boundary_cost_weight = th.ScaleCostWeight(torch.tensor(100.0))

        # --------------------------------------------------------------------------- #
        # -------------------------- Optimization variables ------------------------- #
        # --------------------------------------------------------------------------- #
        # The optimization variables for the motion planer are 2-D positions and
        # velocities for each of the discrete time steps
        poses = []
        velocities = []
        for i in range(self.trajectory_len):
            poses.append(th.Point2(name=f"pose_{i}", dtype=torch.double))
            velocities.append(th.Point2(name=f"vel_{i}", dtype=torch.double))

        # --------------------------------------------------------------------------- #
        # ------------------------------ Cost functions ----------------------------- #
        # --------------------------------------------------------------------------- #
        # Create a Theseus objective for adding the cost functions
        objective = th.Objective(dtype=torch.double)

        # First create the cost functions for the end point positions and velocities
        # which are hard constraints, and can be implemented via Difference cost
        # functions.
        objective.add(
            th.Difference(poses[0], start_point, boundary_cost_weight, name="pose_0")
        )
        objective.add(
            th.Difference(
                velocities[0],
                th.Point2(tensor=torch.zeros(1, 2)),
                boundary_cost_weight,
                name="vel_0",
            )
        )
        objective.add(
            th.Difference(poses[-1], goal_point, boundary_cost_weight, name="pose_N")
        )
        objective.add(
            th.Difference(
                velocities[-1],
                th.Point2(tensor=torch.zeros(1, 2)),
                boundary_cost_weight,
                name="vel_N",
            )
        )

        # Next we add 2-D collisions and GP cost functions, and associate them with the
        # cost weights created above. We need a separate cost function for each time
        # step
        for i in range(1, self.trajectory_len):
            objective.add(
                th.eb.Collision2D(
                    poses[i],
                    sdf_origin,
                    sdf_data_tensor,
                    cell_size_tensor,
                    cost_eps,
                    collision_cost_weights[0]
                    if use_single_collision_weight
                    else collision_cost_weights[i - 1],
                    name=f"collision_{i}",
                )
            )
            objective.add(
                (
                    th.eb.GPMotionModel(
                        poses[i - 1],
                        velocities[i - 1],
                        poses[i],
                        velocities[i],
                        dt,
                        gp_cost_weight,
                        name=f"gp_{i}",
                    )
                )
            )

        # Finally, create the Nonlinear Least Squares optimizer for this objective
        # and wrap both into a TheseusLayer
        optimizer: th.NonlinearLeastSquares
        if optim_method == "gauss_newton":
            optimizer = th.GaussNewton(
                objective,
                th.CholeskyDenseSolver,
                max_iterations=max_optim_iters,
                step_size=step_size,
            )
        elif optim_method == "levenberg_marquardt":
            optimizer = th.LevenbergMarquardt(
                objective,
                th.CholeskyDenseSolver,
                max_iterations=max_optim_iters,
                step_size=step_size,
            )

        self.objective = objective
        self.layer = th.TheseusLayer(optimizer)
        self.layer.to(device=device, dtype=dtype)

        # A call to motion_planner.layer.forward(input_dict) will run the NLLS optimizer
        # to solve for a trajectory given the input data. The input dictionary, supports
        # the keys and values illustrated below, where ':' represents a batch dimension.
        #
        # input_dict = {
        #   "cell_size": tensor(:, 1),
        #   "cost_eps": tensor(:, 1),
        #   "dt": tensor(:, 1),
        #   "sdf_origin": tensor (:, 2),
        #   "start": tensor (:, 2),
        #   "goal": tensor(:, 2),
        #   "sdf_data": tensor(:, map_size, map_size),
        #   "collision_w_0": tensor(:, 1),
        #   "collision_w_1": tensor(:, 1),
        #   ...
        #   "pose_0": tensor(:, 2),
        #   "pose_1": tensor(:, 2),
        #   ...
        #   "pose_N": tensor(:, 2),
        #   "vel_0": tensor(:, 2),
        #   "vel_1": tensor(:, 2),
        #   ...
        #   "vel_N": tensor(:, 2),
        # }
        #
        # TheseusLayer will match the names to their corresponding variables and update
        # their data before calling the optimizer.
        #
        # **Important**, all of these keys are
        # optional, so that only variables that need to change from call to call need to
        # have their names passed. Those not updated explictly here will retain their
        # previous internal data.
        #
        # When running
        #   output, info = motion_planner.forward(input_dict)
        #
        # The motion planner will only modify optimization variables, and all other
        # variables are not modified. For convenience, the output is a dictionary of
        # (str, tensor) mapping variable names to optimized variable data tensors.

    def get_variable_values_from_straight_line(
        self, start: torch.Tensor, goal: torch.Tensor
    ) -> Dict[str, torch.Tensor]:
        # Returns a dictionary of variable names to values that represent a straight
        # line trajectory from start to goal.
        start_goal_dist = goal - start
        avg_vel = start_goal_dist / self.total_time
        unit_trajectory_len = start_goal_dist / (self.trajectory_len - 1)
        input_dict: Dict[str, torch.Tensor] = {}
        for i in range(self.trajectory_len):
            input_dict[f"pose_{i}"] = start + unit_trajectory_len * i
            input_dict[f"vel_{i}"] = avg_vel
        return input_dict

    def get_random_variable_values(
        self, start: torch.Tensor
    ) -> Dict[str, torch.Tensor]:
        # Returns a dictionary of variable names with random initial poses.
        input_dict: Dict[str, torch.Tensor] = {}
        for i in range(self.trajectory_len):
            input_dict[f"pose_{i}"] = torch.randn_like(start)
            input_dict[f"vel_{i}"] = torch.randn_like(start)
        return input_dict

    def get_variable_values_from_trajectory(
        self, trajectory: torch.Tensor
    ) -> Dict[str, torch.Tensor]:
        # Returns a dictionary of variable names to values, so that values
        # are assigned with the data from the given trajectory. Trajectory should be a
        # tensor of shape (batch_size, 4, planner.trajectory_len).
        assert trajectory.shape[1:] == (4, self.trajectory_len)
        input_dict: Dict[str, torch.Tensor] = {}
        for i in range(self.trajectory_len):
            input_dict[f"pose_{i}"] = trajectory[:, :2, i]
            input_dict[f"vel_{i}"] = trajectory[:, :2, i]
        return input_dict

    def error(self) -> float:
        # Returns the current MSE of the optimization problem
        with torch.no_grad():
            return self.objective.error_squared_norm().mean().item()

    def get_trajectory(
        self,
        values_dict: Optional[Dict[str, torch.Tensor]] = None,
        detach: bool = False,
    ) -> torch.Tensor:
        # Returns the a tensor with the trajectory that the given variable
        # values represent. If no dictionary is passed, it will used the latest
        # values stored in the objective's variables.
        trajectory = torch.empty(
            self.objective.batch_size,
            4,
            self.trajectory_len,
            device=self.objective.device,
        )
        variables = self.objective.optim_vars
        for i in range(self.trajectory_len):
            if values_dict is None:
                trajectory[:, :2, i] = variables[f"pose_{i}"].tensor.clone()
                trajectory[:, 2:, i] = variables[f"vel_{i}"].tensor.clone()
            else:
                trajectory[:, :2, i] = values_dict[f"pose_{i}"]
                trajectory[:, 2:, i] = values_dict[f"vel_{i}"]
        return trajectory.detach() if detach else trajectory

    def get_total_squared_errors(self) -> Tuple[torch.Tensor, torch.Tensor]:
        gp_error: torch.Tensor = 0  # type: ignore
        collision_error: torch.Tensor = 0  # type: ignore
        for name, cf in self.objective.cost_functions.items():
            if "gp" in name:
                gp_error += cf.error().square().mean()
            if "collision" in name:
                collision_error += cf.error().square().mean()
        return gp_error, collision_error

    def copy(self, collision_weight: Optional[float] = None) -> "MotionPlanner":
        return MotionPlanner(
            self.map_size,
            self.epsilon_dist,
            self.total_time,
            collision_weight or self.collision_weight,
            self.Qc_inv,
            self.num_time_steps,
            self.optim_method,
            self.max_optim_iters,
            step_size=self.step_size,
            use_single_collision_weight=self.use_single_collision_weight,
            device=self.device,
            dtype=self.dtype,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Optional, Union, cast

import torch

import theseus.constants

from .lie_group import LieGroup
from .point_types import Point2
from .so2 import SO2


# If tensor is passed, must be x, y, cos, sin
# If x_y_theta is passed, must be tensor with shape batch_size x 3
class SE2(LieGroup):
    def __init__(
        self,
        x_y_theta: Optional[torch.Tensor] = None,
        tensor: Optional[torch.Tensor] = None,
        name: Optional[str] = None,
        dtype: Optional[torch.dtype] = None,
        strict: bool = False,
    ):
        if x_y_theta is not None and tensor is not None:
            raise ValueError("Please provide only one of x_y_theta or tensor.")
        if x_y_theta is not None:
            dtype = x_y_theta.dtype
        super().__init__(tensor=tensor, name=name, dtype=dtype, strict=strict)
        if x_y_theta is not None:
            self.update_from_x_y_theta(x_y_theta)

        self._resolve_eps()

    def _resolve_eps(self):
        self._NEAR_ZERO_EPS = theseus.constants._SE2_NEAR_ZERO_EPS[self.tensor.dtype]

    @staticmethod
    def rand(
        *size: int,
        generator: Optional[torch.Generator] = None,
        dtype: Optional[torch.dtype] = None,
        device: Optional[torch.device] = None,
        requires_grad: bool = False,
    ) -> "SE2":
        if len(size) != 1:
            raise ValueError("The size should be 1D.")
        x_y_theta = torch.rand(
            size[0],
            3,
            generator=generator,
            dtype=dtype,
            device=device,
            requires_grad=requires_grad,
        )
        x_y_theta[:, 2] = 2 * theseus.constants.PI * (x_y_theta[:, 2] - 0.5)

        return SE2(x_y_theta=x_y_theta)

    @staticmethod
    def randn(
        *size: int,
        generator: Optional[torch.Generator] = None,
        dtype: Optional[torch.dtype] = None,
        device: Optional[torch.device] = None,
        requires_grad: bool = False,
    ) -> "SE2":
        if len(size) != 1:
            raise ValueError("The size should be 1D.")
        x_y_theta = torch.randn(
            size[0],
            3,
            generator=generator,
            dtype=dtype,
            device=device,
            requires_grad=requires_grad,
        )
        x_y_theta[:, 2] *= theseus.constants.PI

        return SE2(x_y_theta=x_y_theta)

    def _transform_shape_check(self, point: Union[Point2, torch.Tensor]):
        err_msg = (
            f"SE2 can only transform vectors of shape [{self.shape[0]}, 2] or [1, 2], "
            f"but the input has shape {point.shape}."
        )

        if isinstance(point, torch.Tensor):
            if not point.ndim == 2 or point.shape[1] != 2:
                raise ValueError(err_msg)
        elif point.dof() != 2:
            raise ValueError(err_msg)
        if (
            point.shape[0] != self.shape[0]
            and point.shape[0] != 1
            and self.shape[0] != 1
        ):
            raise ValueError(
                "Input point batch size is not broadcastable with group batch size."
            )

    @staticmethod
    def _init_tensor() -> torch.Tensor:  # type: ignore
        return torch.tensor([0.0, 0.0, 1.0, 0.0]).view(1, 4)

    def dof(self) -> int:
        return 3

    def __repr__(self) -> str:
        return f"SE2(tensor={self.tensor}, name={self.name})"

    def __str__(self) -> str:
        with torch.no_grad():
            theta = torch.atan2(self[:, 3:], self[:, 2:3])
            xytheta = torch.cat([self[:, :2], theta], dim=1)
            return f"SE2(xytheta={xytheta}, name={self.name})"

    @property
    def rotation(self) -> SO2:
        return SO2(tensor=self[:, 2:])

    def theta(self, jacobians: Optional[List[torch.Tensor]] = None) -> torch.Tensor:
        if jacobians is not None:
            self._check_jacobians_list(jacobians)
            J_out = torch.zeros(
                self.shape[0], 1, 3, device=self.device, dtype=self.dtype
            )
            J_out[:, 0, 2] = 1
            jacobians.append(J_out)
        return self.rotation.theta()

    @property
    def translation(self) -> Point2:
        return Point2(tensor=self[:, :2])

    def xy(self, jacobians: Optional[List[torch.Tensor]] = None) -> torch.Tensor:
        if jacobians is not None:
            self._check_jacobians_list(jacobians)
            rotation = self.rotation
            J_out = torch.zeros(
                self.shape[0], 2, 3, device=self.device, dtype=self.dtype
            )
            J_out[:, :2, :2] = rotation.to_matrix()
            jacobians.append(J_out)
        return self.translation.tensor

    def update_from_x_y_theta(self, x_y_theta: torch.Tensor):
        rotation = SO2(theta=x_y_theta[:, 2:])
        translation = Point2(tensor=x_y_theta[:, :2])
        self.update_from_rot_and_trans(rotation, translation)

    def update_from_rot_and_trans(self, rotation: SO2, translation: Point2):
        batch_size = rotation.shape[0]
        self.tensor = torch.empty(batch_size, 4).to(
            device=rotation.device, dtype=rotation.dtype
        )
        self[:, :2] = translation.tensor
        cosine, sine = rotation.to_cos_sin()
        self[:, 2] = cosine
        self[:, 3] = sine

    def _log_map_impl(
        self, jacobians: Optional[List[torch.Tensor]] = None
    ) -> torch.Tensor:
        rotation = self.rotation
        theta = rotation.log_map().view(-1)
        cosine, sine = rotation.to_cos_sin()

        # Compute the approximations when theta is near to 0
        small_theta = theta.abs() < self._NEAR_ZERO_EPS
        non_zero = torch.ones(1, dtype=self.dtype, device=self.device)
        sine_nz = torch.where(small_theta, non_zero, sine)
        half_theta_by_tan_half_theta = (
            0.5
            * (1 + cosine)
            * torch.where(small_theta, 1 + sine**2 / 6, theta / sine_nz)
        )
        half_theta = 0.5 * theta

        # Compute the translation
        ux = half_theta_by_tan_half_theta * self[:, 0] + half_theta * self[:, 1]
        uy = half_theta_by_tan_half_theta * self[:, 1] - half_theta * self[:, 0]

        if jacobians is not None:
            SE2._check_jacobians_list(jacobians)
            jac = torch.zeros(
                self.shape[0],
                3,
                3,
                dtype=self.dtype,
                device=self.device,
            )

            theta2 = theta**2
            theta3 = theta * theta2

            theta_nz = torch.where(small_theta, non_zero, theta)
            one_minus_cosine_nz = torch.where(small_theta, non_zero, 1 - cosine)

            half_theta_sine_by_one_minus_cosine = torch.where(
                small_theta, 1 - theta2 / 12.0, half_theta * sine / one_minus_cosine_nz
            )
            jac[:, [0, 1], [0, 1]] = half_theta_sine_by_one_minus_cosine.view(-1, 1)
            jac[:, 0, 1] = -half_theta
            jac[:, 1, 0] = half_theta

            coeff = torch.where(
                small_theta,
                theta / 12.0 + theta3 / 720.0,
                1.0 / theta_nz - 0.5 * sine / one_minus_cosine_nz,
            )

            jac[:, 0, 2] = coeff * ux + 0.5 * uy
            jac[:, 1, 2] = coeff * uy - 0.5 * ux

            jac[:, 2, 2] = 1

            jacobians.append(jac)

        return torch.stack((ux, uy, theta), dim=1)

    @staticmethod
    def _check_tensor_impl(tensor: torch.Tensor):
        with torch.no_grad():
            if tensor.ndim != 2 or tensor.shape[1] != 4:
                raise ValueError("SE2 data tensors can only be 4D vectors.")

            return SO2._check_tensor_impl(tensor[:, 2:])

    @staticmethod
    def exp_map(
        tangent_vector: torch.Tensor, jacobians: Optional[List[torch.Tensor]] = None
    ) -> "SE2":
        u = tangent_vector[:, :2]
        theta = tangent_vector[:, 2]
        rotation = SO2(theta=theta)

        cosine, sine = rotation.to_cos_sin()

        # Compute the approximations when theta is near to 0
        small_theta = (
            theta.abs() < theseus.constants._SE2_NEAR_ZERO_EPS[tangent_vector.dtype]
        )
        non_zero = torch.ones(
            1, dtype=tangent_vector.dtype, device=tangent_vector.device
        )
        theta2 = theta**2
        theta3 = theta**3
        theta_nz = torch.where(small_theta, non_zero, theta)
        sine_by_theta = torch.where(small_theta, 1 - theta2 / 6, sine / theta_nz)
        cosine_minus_one_by_theta = torch.where(
            small_theta, -theta / 2 + theta3 / 24, (cosine - 1) / theta_nz
        )

        # Compute the translation
        x = sine_by_theta * u[:, 0] + cosine_minus_one_by_theta * u[:, 1]
        y = sine_by_theta * u[:, 1] - cosine_minus_one_by_theta * u[:, 0]
        translation = Point2(tensor=torch.stack((x, y), dim=1))

        se2 = SE2(dtype=tangent_vector.dtype)
        se2.update_from_rot_and_trans(rotation, translation)

        if jacobians is not None:
            SE2._check_jacobians_list(jacobians)
            theta2_nz = torch.where(small_theta, non_zero, theta2)
            theta_minus_sine_by_theta2 = torch.where(
                small_theta, theta - theta3 / 120, (theta - sine) / theta2_nz
            )
            cosine_minus_one_by_theta2 = torch.where(
                small_theta, -0.5 + theta2 / 24, (cosine - 1) / theta2_nz
            )
            jac = torch.zeros(
                tangent_vector.shape[0],
                3,
                3,
                dtype=tangent_vector.dtype,
                device=tangent_vector.device,
            )
            jac[:, 0, 0] = sine_by_theta
            jac[:, 0, 1] = -cosine_minus_one_by_theta
            jac[:, 0, 2] = (
                theta_minus_sine_by_theta2 * u[:, 0]
                + cosine_minus_one_by_theta2 * u[:, 1]
            )
            jac[:, 1, 0] = cosine_minus_one_by_theta
            jac[:, 1, 1] = sine_by_theta
            jac[:, 1, 2] = (
                theta_minus_sine_by_theta2 * u[:, 1]
                - cosine_minus_one_by_theta2 * u[:, 0]
            )
            jac[:, 2, 2] = 1
            jacobians.append(jac)

        return se2

    @staticmethod
    def normalize(tensor: torch.Tensor) -> torch.Tensor:
        if tensor.ndim != 2 or tensor.shape[1] != 4:
            raise ValueError("SE2 data tensors can only be 4D vectors.")

        return torch.cat([tensor[:, :2], SO2.normalize(tensor[:, 2:])], dim=1)

    def _adjoint_impl(self) -> torch.Tensor:
        ret = torch.zeros(self.shape[0], 3, 3).to(device=self.device, dtype=self.dtype)
        ret[:, :2, :2] = self.rotation.to_matrix()
        ret[:, 0, 2] = self[:, 1]
        ret[:, 1, 2] = -self[:, 0]
        ret[:, 2, 0] = 0
        ret[:, 2, 2] = 1
        return ret

    def _compose_impl(self, se2_2: LieGroup) -> "SE2":
        se2_2 = cast(SE2, se2_2)
        rotation_1 = self.rotation
        rotation_2 = se2_2.rotation
        translation_1 = cast(Point2, self.translation)
        translation_2 = cast(Point2, se2_2.translation)
        new_rotation = rotation_1.compose(rotation_2)
        new_translation = cast(
            Point2,
            translation_1.compose(rotation_1.rotate(translation_2)),
        )
        return SE2(
            tensor=torch.cat([new_translation.tensor, new_rotation.tensor], dim=1),
            strict=False,
        )

    def _inverse_impl(self) -> "SE2":
        inverse_rotation = self.rotation._inverse_impl()
        inverse_translation = inverse_rotation.rotate(cast(Point2, -self.translation))
        se2_inverse = SE2(dtype=self.dtype)
        se2_inverse.update_from_rot_and_trans(inverse_rotation, inverse_translation)
        return se2_inverse

    def _project_impl(
        self, euclidean_grad: torch.Tensor, is_sparse: bool = False
    ) -> torch.Tensor:
        self._project_check(euclidean_grad, is_sparse)
        ret = torch.zeros(
            euclidean_grad.shape[:-1] + torch.Size([3]),
            dtype=self.dtype,
            device=self.device,
        )

        temp = torch.stack((-self[:, 3], self[:, 2]), dim=1)

        if is_sparse:
            ret[..., 0] = torch.einsum(
                "i...k,i...k->i...", euclidean_grad[..., :2], self[:, 2:]
            )
            ret[..., 1] = torch.einsum(
                "i...k,i...k->i...", euclidean_grad[..., :2], temp
            )
            ret[..., 2] = torch.einsum(
                "i...k,i...k->i...", euclidean_grad[..., 2:], temp
            )
        else:
            ret[..., 0] = torch.einsum(
                "...k,...k", euclidean_grad[..., :2], self[:, 2:]
            )
            ret[..., 1] = torch.einsum("...k,...k", euclidean_grad[..., :2], temp)
            ret[..., 2] = torch.einsum("...k,...k", euclidean_grad[..., 2:], temp)
        return ret

    def to_matrix(self) -> torch.Tensor:
        matrix = torch.zeros(self.shape[0], 3, 3).to(
            device=self.device, dtype=self.dtype
        )
        matrix[:, :2, :2] = self.rotation.to_matrix()
        matrix[:, :2, 2] = self[:, :2]
        matrix[:, 2, 2] = 1.0
        return matrix

    @staticmethod
    def hat(tangent_vector: torch.Tensor) -> torch.Tensor:
        theta = tangent_vector[:, 2]
        u = tangent_vector[:, :2]
        matrix = torch.zeros(tangent_vector.shape[0], 3, 3).to(
            device=tangent_vector.device, dtype=tangent_vector.dtype
        )
        matrix[:, 0, 1] = -theta
        matrix[:, 1, 0] = theta
        matrix[:, :2, 2] = u
        return matrix

    @staticmethod
    def vee(matrix: torch.Tensor) -> torch.Tensor:
        _check = matrix.ndim == 3 and matrix.shape[1:] == (3, 3)
        _check &= torch.allclose(matrix[:, 0, 1], -matrix[:, 1, 0])
        _check &= matrix[:, 0, 0].abs().max().item() < theseus.constants.EPS
        _check &= matrix[:, 1, 1].abs().max().item() < theseus.constants.EPS
        _check &= matrix[:, 2, 2].abs().max().item() < theseus.constants.EPS
        if not _check:
            raise ValueError("Invalid hat matrix for SE2.")
        batch_size = matrix.shape[0]
        tangent_vector = torch.zeros(batch_size, 3).to(
            device=matrix.device, dtype=matrix.dtype
        )
        tangent_vector[:, 2] = matrix[:, 1, 0]
        tangent_vector[:, :2] = matrix[:, :2, 2]
        return tangent_vector

    def transform_to(
        self,
        point: Union[torch.Tensor, Point2],
        jacobians: Optional[List[torch.Tensor]] = None,
    ) -> Point2:
        self._transform_shape_check(point)
        batch_size = max(self.shape[0], point.shape[0])
        if isinstance(point, torch.Tensor):
            p = point
        else:
            p = point.tensor

        cosine = self[:, 2]
        sine = self[:, 3]
        temp = p - self[:, :2]
        ret = SO2._rotate_from_cos_sin(temp, cosine, -sine)

        if jacobians is not None:
            self._check_jacobians_list(jacobians)
            Jg = torch.zeros(batch_size, 2, 3, dtype=self.dtype, device=self.device)
            Jg[:, 0, 0] = -1
            Jg[:, 1, 1] = -1
            Jg[:, 0, 2] = ret.y()
            Jg[:, 1, 2] = -ret.x()

            Jpnt = torch.zeros(batch_size, 2, 2, dtype=self.dtype, device=self.device)
            Jpnt[:, 0, 0] = cosine
            Jpnt[:, 0, 1] = sine
            Jpnt[:, 1, 0] = -sine
            Jpnt[:, 1, 1] = cosine

            jacobians.extend([Jg, Jpnt])

        return ret

    def transform_from(
        self,
        point: Union[torch.Tensor, Point2],
        jacobians: Optional[List[torch.Tensor]] = None,
    ) -> Point2:
        self._transform_shape_check(point)
        batch_size = max(self.shape[0], point.shape[0])

        cosine = self[:, 2]
        sine = self[:, 3]
        temp = SO2._rotate_from_cos_sin(point, cosine, sine)
        ret = Point2(tensor=temp.tensor + self[:, :2])

        if jacobians is not None:
            self._check_jacobians_list(jacobians)
            Jg = torch.zeros(batch_size, 2, 3, dtype=self.dtype, device=self.device)
            Jg[:, 0, 0] = cosine
            Jg[:, 0, 1] = -sine
            Jg[:, 1, 0] = sine
            Jg[:, 1, 1] = cosine
            Jg[:, 0, 2] = -temp.y()
            Jg[:, 1, 2] = temp.x()

            Jpnt = torch.zeros(batch_size, 2, 2, dtype=self.dtype, device=self.device)
            Jpnt[:, 0, 0] = cosine
            Jpnt[:, 0, 1] = -sine
            Jpnt[:, 1, 0] = sine
            Jpnt[:, 1, 1] = cosine

            jacobians.extend([Jg, Jpnt])

        return ret

    def _copy_impl(self, new_name: Optional[str] = None) -> "SE2":
        return SE2(tensor=self.tensor.clone(), name=new_name)

    # only added to avoid casting downstream
    def copy(self, new_name: Optional[str] = None) -> "SE2":
        return cast(SE2, super().copy(new_name=new_name))

    # calls to() on the internal tensors
    def to(self, *args, **kwargs):
        super().to(*args, **kwargs)
        self._resolve_eps()


rand_se2 = SE2.rand
randn_se2 = SE2.randn



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from itertools import count
from typing import Optional, Union

import torch


class Variable:
    """A variable in a differentiable optimization problem."""

    _ids = count(0)

    def __init__(self, tensor: torch.Tensor, name: Optional[str] = None):
        self._id = next(Variable._ids)
        self._num_updates = 0
        if name:
            self.name = name
        else:
            self.name = f"{self.__class__.__name__}__{self._id}"
        self.tensor = tensor

    def copy(self, new_name: Optional[str] = None) -> "Variable":
        if not new_name:
            new_name = f"{self.name}_copy"
        return Variable(self.tensor.clone(), name=new_name)

    def __deepcopy__(self, memo):
        if id(self) in memo:
            return memo[id(self)]
        the_copy = self.copy()
        memo[id(self)] = the_copy
        return the_copy

    # batch_ignore_mask is a boolean list where batch_ignore_mask[i] = 1 means
    # variable[i] will *not* be updated
    def update(
        self,
        data: Union[torch.Tensor, "Variable"],
        batch_ignore_mask: Optional[torch.Tensor] = None,
    ):
        if isinstance(data, Variable):
            tensor = data.tensor
        else:
            tensor = data
        if (
            len(tensor.shape) != len(self.tensor.shape)
            or tensor.shape[1:] != self.tensor.shape[1:]
        ):
            raise ValueError(
                f"Tried to update tensor {self.name} with data "
                f"incompatible with original tensor shape. Given {tensor.shape[1:]}. "
                f"Expected: {self.tensor.shape[1:]}"
            )
        if tensor.dtype != self.dtype:
            raise ValueError(
                f"Tried to update used tensor of dtype {tensor.dtype} but Variable "
                f"{self.name} has dtype {self.dtype}."
            )
        if batch_ignore_mask is not None and batch_ignore_mask.any():
            mask_shape = (-1,) + (1,) * (tensor.ndim - 1)
            self.tensor = torch.where(
                batch_ignore_mask.view(mask_shape), self.tensor, tensor
            )
        else:
            self.tensor = tensor
        self._num_updates += 1

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(tensor={self.tensor}, name={self.name})"

    def __str__(self) -> str:
        return repr(self)

    # calls to() on the internal tensors
    def to(self, *args, **kwargs):
        self.tensor = self.tensor.to(*args, **kwargs)

    @property
    def shape(self) -> torch.Size:
        return self.tensor.shape

    @property
    def device(self) -> torch.device:
        return self.tensor.device

    @property
    def dtype(self) -> torch.dtype:
        return self.tensor.dtype

    @property
    def ndim(self) -> int:
        return self.tensor.ndim

    def __getitem__(self, item):
        return self.tensor[item]

    def __setitem__(self, item, value):
        self.tensor[item] = value



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Callable, List, Optional, Type

import torch
import torch.nn as nn

import theseus.geometry as th


def build_mlp(
    input_dim: int,
    hidden_dim: int,
    output_dim: int,
    hidden_depth: int,
    act_name: str = "relu",
) -> torch.nn.Sequential:
    # assign types
    act: Type[nn.Module]
    mods: List[nn.Module]

    # find activation
    if act_name == "relu":
        act = nn.ReLU
    elif act_name == "elu":
        act = nn.ELU
    elif act_name == "tanh":
        act = nn.Tanh
    else:
        raise NotImplementedError()

    # construct sequential list of modules
    if hidden_depth == 0:
        mods = [nn.Linear(input_dim, output_dim)]
    else:
        mods = [nn.Linear(input_dim, hidden_dim), act()]
        for i in range(hidden_depth - 1):
            mods += [nn.Linear(hidden_dim, hidden_dim), act()]
        mods.append(nn.Linear(hidden_dim, output_dim))
    trunk = nn.Sequential(*mods)
    return trunk


# Inputs are a batched matrix and batched rows/cols for indexing
# Sizes are:
#      matrix -> batch_size x num_rows x num_cols
#      rows/cols -> batch_size x num_points
#
# If the matrix's batch size is 1, broadcasting is supported by expanding the matrix
# so that it has the batch size of rows and cols.
#
# Returns:
#      data -> batch_size x num_points
# where
#      data[i, j] = matrix[i, rows[i, j], cols[i, j]]
def gather_from_rows_cols(matrix: torch.Tensor, rows: torch.Tensor, cols: torch.Tensor):
    if matrix.shape[0] == 1:
        matrix = matrix.expand(rows.shape[0], -1, -1)
    assert matrix.ndim == 3 and rows.ndim == 2 and rows.ndim == 2
    assert matrix.shape[0] == rows.shape[0] and matrix.shape[0] == cols.shape[0]
    assert rows.shape[1] == cols.shape[1]
    assert rows.min() >= 0 and rows.max() < matrix.shape[1]
    assert cols.min() >= 0 and cols.max() < matrix.shape[2]

    aux_idx = torch.arange(matrix.shape[0]).unsqueeze(-1).to(matrix.device)
    return matrix[aux_idx, rows, cols]


# Computes jacobians of a function of LieGroup arguments using finite differences
#   function: a callable representing the function whose jacobians will be computed
#   group_args: the group values at which the jacobians should be computed
#   function_dim: the dimension of the function. If None, it uses the dimension of the
#      group args.
#   delta_mag: magnitude of the differences used to compute the jacobian
def numeric_jacobian(
    function: Callable[[List[th.LieGroup]], th.LieGroup],
    group_args: List[th.LieGroup],
    function_dim: Optional[int] = None,
    delta_mag: float = 1e-3,
):
    batch_size = max([group_arg.shape[0] for group_arg in group_args])

    def _compute(group_idx):
        dof = group_args[group_idx].dof()
        function_dim_ = function_dim or dof
        jac = torch.zeros(
            batch_size, function_dim_, dof, dtype=group_args[group_idx].dtype
        )
        for d in range(dof):
            delta = torch.zeros(1, dof).to(
                device=group_args[0].device, dtype=group_args[group_idx].dtype
            )
            delta[:, d] = delta_mag

            group_plus = group_args[group_idx].retract(delta)
            group_minus = group_args[group_idx].retract(-delta)
            group_plus_args = [g for g in group_args]
            group_plus_args[group_idx] = group_plus
            group_minus_args = [g for g in group_args]
            group_minus_args[group_idx] = group_minus

            diff = function(group_minus_args).local(function(group_plus_args))
            jac[:, :, d] = diff / (2 * delta_mag)
        return jac

    jacs = []
    for group_idx in range(len(group_args)):
        jacs.append(_compute(group_idx))
    return jacs



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from itertools import count
from typing import List, Optional, Sequence, Tuple

import torch

from theseus.geometry import LieGroup, Manifold


class ManifoldGaussian:
    _ids = count(0)

    def __init__(
        self,
        mean: Sequence[Manifold],
        precision: Optional[torch.Tensor] = None,
        name: Optional[str] = None,
    ):
        self._id = next(ManifoldGaussian._ids)
        if name is None:
            name = f"{self.__class__.__name__}__{self._id}"
        self.name = name

        dof = 0
        for v in mean:
            dof += v.dof()
        self._dof = dof

        self.mean = mean
        if precision is None:
            precision = torch.eye(self.dof).to(
                dtype=mean[0].dtype, device=mean[0].device
            )
            precision = precision[None, ...].repeat(mean[0].shape[0], 1, 1)
        self.update(mean, precision)

    @property
    def dof(self) -> int:
        return self._dof

    @property
    def device(self) -> torch.device:
        return self.mean[0].device

    @property
    def dtype(self) -> torch.dtype:
        return self.mean[0].dtype

    # calls to() on the internal tensors
    def to(self, *args, **kwargs):
        for var in self.mean:
            var = var.to(*args, **kwargs)
        self.precision = self.precision.to(*args, **kwargs)

    def copy(self, new_name: Optional[str] = None) -> "ManifoldGaussian":
        if not new_name:
            new_name = f"{self.name}_copy"
        mean_copy = [var.copy() for var in self.mean]
        precision_copy = self.precision.clone()
        return ManifoldGaussian(mean_copy, precision=precision_copy, name=new_name)

    def __deepcopy__(self, memo):
        if id(self) in memo:
            return memo[id(self)]
        the_copy = self.copy()
        memo[id(self)] = the_copy
        return the_copy

    def update(
        self,
        mean: Sequence[Manifold],
        precision: torch.Tensor,
    ):
        if len(mean) != len(self.mean):
            raise ValueError(
                f"Tried to update mean with sequence of different"
                f"length to original mean sequence. Given: {len(mean)}. "
                f"Expected: {len(self.mean)}"
            )
        for i in range(len(self.mean)):
            self.mean[i].update(mean[i])

        expected_shape = torch.Size([mean[0].shape[0], self.dof, self.dof])
        if precision.shape != expected_shape:
            raise ValueError(
                f"Tried to update precision with data "
                f"incompatible with original tensor shape. Given: {precision.shape}. "
                f"Expected: {expected_shape}"
            )
        if precision.dtype != self.dtype:
            raise ValueError(
                f"Tried to update using tensor of dtype: {precision.dtype} but precision "
                f"has dtype: {self.dtype}."
            )
        if precision.device != self.device:
            raise ValueError(
                f"Tried to update using tensor on device: {precision.dtype} but precision "
                f"is on device: {self.device}."
            )
        if not torch.allclose(precision, precision.transpose(1, 2)):
            raise ValueError("Tried to update precision with non-symmetric matrix.")

        self.precision = precision


# Projects the gaussian (ManifoldGaussian object) into the tangent plane at
# variable. The gaussian mean is projected using the local function,
# and the precision is approximately transformed using the jacobains of the exp_map.
# Either returns the mean and precision of the new Gaussian in the tangent plane if
# return_mean is True. Otherwise returns the information vector (eta) and precision.
# See section H, eqn 55 in https://arxiv.org/pdf/1812.01537.pdf for a derivation
# of covariance propagation in manifolds.
def local_gaussian(
    variable: LieGroup,
    gaussian: ManifoldGaussian,
    return_mean: bool = True,
) -> Tuple[torch.Tensor, torch.Tensor]:
    # assumes gaussian is over just one Manifold object
    if len(gaussian.mean) != 1:
        raise ValueError(
            "local on ManifoldGaussian should be over just one Manifold object. "
            f"Passed gaussian {gaussian.name} is over {len(gaussian.mean)} "
            "Manifold objects."
        )
    # check variable and gaussian are of the same LieGroup class
    if gaussian.mean[0].__class__ != variable.__class__:
        raise ValueError(
            "variable and gaussian mean must be instances of the same class. "
            f"variable is of class {variable.__class__} and gaussian mean is "
            f"of class {gaussian.mean[0].__class__}."
        )

    # mean vector in the tangent space at variable
    mean_tp = variable.local(gaussian.mean[0])

    jac: List[torch.Tensor] = []
    variable.exp_map(mean_tp, jacobians=jac)
    # precision matrix in the tangent space at variable
    lam_tp = torch.bmm(torch.bmm(jac[0].transpose(-1, -2), gaussian.precision), jac[0])

    if return_mean:
        return mean_tp, lam_tp
    else:
        eta_tp = torch.matmul(lam_tp, mean_tp.unsqueeze(-1)).squeeze(-1)
        return eta_tp, lam_tp


# Computes the ManifoldGaussian that corresponds to the gaussian in the tangent plane
# at variable, parameterised by the mean (mean_tp) and precision (precision_tp).
# The mean is transformed to a LieGroup element by retraction.
# The precision is transformed using the inverse of the exp_map jacobians.
# See section H, eqn 55 in https://arxiv.org/pdf/1812.01537.pdf for a derivation
# of covariance propagation in manifolds.
def retract_gaussian(
    variable: LieGroup,
    mean_tp: torch.Tensor,
    precision_tp: torch.Tensor,
) -> ManifoldGaussian:
    mean = variable.retract(mean_tp)

    jac: List[torch.Tensor] = []
    variable.exp_map(mean_tp, jacobians=jac)
    inv_jac = torch.inverse(jac[0])
    precision = torch.bmm(torch.bmm(inv_jac.transpose(-1, -2), precision_tp), inv_jac)

    return ManifoldGaussian(mean=[mean], precision=precision)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import pytest  # noqa: F401
import torch  # needed for import of Torch C++ extensions to work
from scipy.sparse import csr_matrix

from theseus.utils import random_sparse_binary_matrix


# ideally we would like to support batch_size <= init_batch_size, but
# because of limitations of cublas those have to be always identical
def check_lu_solver(
    init_batch_size, batch_size, num_rows, num_cols, fill, verbose=False
):
    # this is necessary assumption, so that the hessian is full rank
    assert num_rows >= num_cols

    if not torch.cuda.is_available():
        return
    from theseus.extlib.cusolver_lu_solver import CusolverLUSolver

    A_skel = random_sparse_binary_matrix(
        num_rows, num_cols, fill, min_entries_per_col=3
    )
    A_num_cols = num_cols
    A_rowPtr = torch.tensor(A_skel.indptr, dtype=torch.int).cuda()
    A_colInd = torch.tensor(A_skel.indices, dtype=torch.int).cuda()
    A_num_rows = A_rowPtr.size(0) - 1
    A_nnz = A_colInd.size(0)
    A_val = torch.rand((batch_size, A_nnz), dtype=torch.double).cuda()
    b = torch.rand((batch_size, A_num_rows), dtype=torch.double).cuda()

    A_csr = [
        csr_matrix(
            (A_val[i].cpu(), A_colInd.cpu(), A_rowPtr.cpu()), (A_num_rows, A_num_cols)
        )
        for i in range(batch_size)
    ]
    if verbose:
        print("A[0]:\n", A_csr[0].todense())
        print("b[0]:\n", b[0])

    AtA_csr = [(a.T @ a).tocsr() for a in A_csr]
    AtA_rowPtr = torch.tensor(AtA_csr[0].indptr).cuda()
    AtA_colInd = torch.tensor(AtA_csr[0].indices).cuda()
    AtA_val = torch.tensor(np.array([m.data for m in AtA_csr])).cuda()
    AtA_num_rows = AtA_rowPtr.size(0) - 1
    AtA_num_cols = AtA_num_rows
    AtA_nnz = AtA_colInd.size(0)  # noqa: F841

    if verbose:
        print("AtA[0]:\n", AtA_csr[0].todense())

    slv = CusolverLUSolver(init_batch_size, AtA_num_cols, AtA_rowPtr, AtA_colInd)
    singularities = slv.factor(AtA_val)

    if verbose:
        print("singularities:", singularities)

    b = torch.rand((batch_size, A_num_rows), dtype=torch.double).cuda()
    Atb = torch.tensor(
        np.array([A_csr[i].T @ b[i].cpu().numpy() for i in range(batch_size)])
    ).cuda()
    if verbose:
        print("Atb[0]:", Atb[0])

    sol = Atb.clone()
    slv.solve(sol)
    if verbose:
        print("x[0]:", sol[0])

    residuals = [
        AtA_csr[i] @ sol[i].cpu().numpy() - Atb[i].cpu().numpy()
        for i in range(batch_size)
    ]
    if verbose:
        print("residual[0]:", residuals[0])

    assert all(np.linalg.norm(res) < 1e-10 for res in residuals)


@pytest.mark.cudaext
def test_lu_solver_1():
    check_lu_solver(init_batch_size=5, batch_size=5, num_rows=50, num_cols=30, fill=0.2)


@pytest.mark.cudaext
def test_lu_solver_2():
    check_lu_solver(
        init_batch_size=5, batch_size=5, num_rows=150, num_cols=60, fill=0.2
    )


@pytest.mark.cudaext
def test_lu_solver_3():
    check_lu_solver(
        init_batch_size=10, batch_size=10, num_rows=300, num_cols=90, fill=0.2
    )


@pytest.mark.cudaext
def test_lu_solver_4():
    check_lu_solver(init_batch_size=5, batch_size=5, num_rows=50, num_cols=30, fill=0.1)


@pytest.mark.cudaext
def test_lu_solver_5():
    check_lu_solver(
        init_batch_size=5, batch_size=5, num_rows=150, num_cols=60, fill=0.1
    )


@pytest.mark.cudaext
def test_lu_solver_6():
    check_lu_solver(
        init_batch_size=10, batch_size=10, num_rows=300, num_cols=90, fill=0.1
    )


# would like to test when irregular batch_size < init_batch_size,
# but this is currently not supported by cublas, maybe in the future
# def test_lu_solver_7():
#     check_lu_solver(init_batch_size=10, batch_size=5, num_rows=150, num_cols=60, fill=0.2)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .lu_cuda_sparse_autograd import LUCudaSolveFunction
from .sparse_autograd import CholmodSolveFunction

__all__ = [
    "CholmodSolveFunction",
    "LUCudaSolveFunction",
]



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import pytest  # noqa: F401
import torch

import theseus as th

from .common import run_nonlinear_least_squares_check


@pytest.fixture
def mock_objective():
    objective = th.Objective()
    v1 = th.Vector(1, name="v1")
    v2 = th.Vector(1, name="v2")
    objective.add(th.Difference(v1, v2, th.ScaleCostWeight(1.0)))
    return objective


def test_levenberg_marquartd():
    for ellipsoidal_damping in [True, False]:
        for damping in [0, 0.001, 0.01, 0.1]:
            run_nonlinear_least_squares_check(
                th.LevenbergMarquardt,
                {
                    "damping": damping,
                    "ellipsoidal_damping": ellipsoidal_damping,
                    "damping_eps": 0.0,
                },
                singular_check=damping < 0.001,
            )


def test_ellipsoidal_damping_compatibility(mock_objective):
    mock_objective.update({"v1": torch.ones(1, 1), "v2": torch.zeros(1, 1)})
    for lsc in [th.LUDenseSolver, th.CholeskyDenseSolver]:
        optimizer = th.LevenbergMarquardt(mock_objective, lsc)
        optimizer.optimize(ellipsoidal_damping=True)
        optimizer.optimize(damping_eps=0.1)

    for lsc in [th.CholmodSparseSolver]:
        optimizer = th.LevenbergMarquardt(mock_objective, lsc)
        with pytest.raises(RuntimeError):
            optimizer.optimize(ellipsoidal_damping=True)
        with pytest.raises(RuntimeError):
            optimizer.optimize(damping_eps=0.1)


@pytest.mark.cudaext
def test_ellipsoidal_damping_compatibility_cuda(mock_objective):
    mock_objective.to(device="cuda", dtype=torch.double)
    batch_size = 2
    mock_objective.update(
        {
            "v1": torch.ones(batch_size, 1, device="cuda", dtype=torch.double),
            "v2": torch.zeros(batch_size, 1, device="cuda", dtype=torch.double),
        }
    )
    for lsc in [th.LUCudaSparseSolver]:
        optimizer = th.LevenbergMarquardt(
            mock_objective, lsc, linear_solver_kwargs={"batch_size": batch_size}
        )
        optimizer.optimize(ellipsoidal_damping=True)
        optimizer.optimize(damping_eps=0.1)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import warnings
from typing import List, Optional, Tuple, Type

import torch

from .cost_function import CostFunction
from .cost_weight import CostWeight
from .robust_loss import RobustLoss
from .variable import Variable


# This class implements a robust cost function that incorporates a loss for
# reducing the influence of outliers.
# See http://ceres-solver.org/nnls_modeling.html#lossfunction
#
# It works by wrapping a CostFunction, inheriting all of its optimization variables
# and auxiliary variables, without copying them. During linearization, the loss is
# applied to the error's squared norm, linearized, and used to rescale the underlying
# cost function's error and jacobians to solve for the full robust cost function
# (see Theory section at http://ceres-solver.org/nnls_modeling.html#theory and
# references therein); here we use alpha=0. The implementation of this part is done via
# `RobustCostFunction.weighted_jacobians_error()`.
#
# Due to the nature of `RobustCostFunction`, its behavior is different to typical cost
# functions. Below we use the notation e, J, w, to refer to the error, jacobian, and
# weight of the base cost function, and let rho be the robust loss used. The important
# points are the following:
#
#   - For h := robust_cost_fn.weighted_error(), we have ||h||2 == rho(||w * e||2)
#   - For r_J, r_e defined as any jacobian/error returned by
#     robust_cost_fn.weighted_jacobians_error(), we have that
#     r_Jv^T * r_e == rho' * J^T * e, which is the gradient of rho(||w * e||2).
#   - Note that h != r_e. In general, weighted_jacobians_and_error()
#       is used by our optimizers, since it allows RobustCostFunction to be coupled
#       with any Jacobian-based NLS solver w/o modifications. However, if you are
#       interested in the robust cost value itself, you should use the error returned
#       by `weighted_error()` and **NOT** the one returned by
#       `weighted_jacobians_error()`.
#
# Finally, since we apply the weight before the robust loss, we adopt the convention
# that `robust_cost_fn.jacobians() == robust_cost_fn.weighted_jacobians_error()`, and
# `robust_cost_fn.error() == robust_cost_fn.weighed_error()`.
class RobustCostFunction(CostFunction):
    _EPS = 1e-20

    def __init__(
        self,
        cost_function: CostFunction,
        loss_cls: Type[RobustLoss],
        log_loss_radius: Variable,
        name: Optional[str] = None,
    ):
        self.cost_function = cost_function
        super().__init__(cost_function.weight, name=name)

        # Register optimization variables of the underlying cost function
        for attr in cost_function._optim_vars_attr_names:
            setattr(self, attr, getattr(cost_function, attr))
            self.register_optim_var(attr)

        # Register auxiliary variables of the underlying cost function
        for attr in cost_function._aux_vars_attr_names:
            setattr(self, attr, getattr(cost_function, attr))
            self.register_aux_var(attr)

        self.log_loss_radius = log_loss_radius
        self.register_aux_var("log_loss_radius")
        self.loss = loss_cls()

    def error(self) -> torch.Tensor:
        warnings.warn(
            "Computing the robust cost error requires weighting first, so "
            "error() is equivalent to weighted_error()."
        )
        return self.weighted_error()

    def weighted_error(self) -> torch.Tensor:
        weighted_error = self.cost_function.weighted_error()
        squared_norm = torch.sum(weighted_error**2, dim=1, keepdim=True)
        error_loss = self.loss.evaluate(squared_norm, self.log_loss_radius.tensor)

        # The return value is a hacky way to make it so that
        # ||weighted_error||^2 = error_loss
        # By doing this we avoid having to change the objective's error computation
        # specifically for robust cost functions. The issue for this type of cost
        # function is that the theory requires us to maintain scaled errors/jacobians
        # of dim = robust_fn.cost_function.dim() to do the linearization properly,
        # but the actual error has dim = 1, being the result of loss(||error||^2).
        return (
            torch.ones_like(weighted_error)
            * (error_loss / self.dim() + RobustCostFunction._EPS).sqrt()
        )

    def jacobians(self) -> Tuple[List[torch.Tensor], torch.Tensor]:
        warnings.warn(
            "Computing the robust cost error requires weighting first, so "
            "jacobians() is equivalent to weighted_jacobians_error()."
        )
        return self.weighted_jacobians_error()

    def weighted_jacobians_error(self) -> Tuple[List[torch.Tensor], torch.Tensor]:
        (
            weighted_jacobians,
            weighted_error,
        ) = self.cost_function.weighted_jacobians_error()
        squared_norm = torch.sum(weighted_error**2, dim=1, keepdim=True)
        rescale = (
            self.loss.linearize(squared_norm, self.log_loss_radius.tensor)
            + RobustCostFunction._EPS
        ).sqrt()

        return [
            rescale.view(-1, 1, 1) * jacobian for jacobian in weighted_jacobians
        ], rescale * weighted_error

    def dim(self) -> int:
        return self.cost_function.dim()

    def _copy_impl(self, new_name: Optional[str] = None) -> "RobustCostFunction":
        return RobustCostFunction(
            self.cost_function.copy(),
            type(self.loss),
            self.log_loss_radius.copy(),
            name=new_name,
        )

    @property
    def weight(self) -> CostWeight:
        return self.cost_function.weight

    @weight.setter
    def weight(self, weight: CostWeight):
        self.cost_function.weight = weight



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Optional, Tuple

import torch

from theseus.core import CostFunction, CostWeight
from theseus.geometry import LieGroup, between


class MovingFrameBetween(CostFunction):
    def __init__(
        self,
        frame1: LieGroup,
        frame2: LieGroup,
        pose1: LieGroup,
        pose2: LieGroup,
        measurement: LieGroup,
        cost_weight: CostWeight,
        name: Optional[str] = None,
    ):
        seen_classes = set(
            [x.__class__.__name__ for x in [frame1, frame2, pose1, pose2, measurement]]
        )
        if len(seen_classes) > 1:
            raise ValueError("Inconsistent types between input variables.")

        super().__init__(cost_weight, name=name)
        self.frame1 = frame1
        self.frame2 = frame2
        self.pose1 = pose1
        self.pose2 = pose2
        self.register_optim_vars(["frame1", "frame2", "pose1", "pose2"])
        self.measurement = measurement
        self.register_aux_vars(["measurement"])

    def error(self) -> torch.Tensor:
        pose1__frame = between(self.frame1, self.pose1)
        pose2__frame = between(self.frame2, self.pose2)
        var_diff = between(pose1__frame, pose2__frame)
        return self.measurement.local(var_diff)

    def jacobians(self) -> Tuple[List[torch.Tensor], torch.Tensor]:
        jacobians_b1: List[torch.Tensor] = []
        pose1__frame = between(self.frame1, self.pose1, jacobians=jacobians_b1)
        jacobians_b2: List[torch.Tensor] = []
        pose2__frame = between(self.frame2, self.pose2, jacobians=jacobians_b2)
        jacobians_b_out: List[torch.Tensor] = []
        var_diff = between(pose1__frame, pose2__frame, jacobians=jacobians_b_out)
        error = self.measurement.local(var_diff)

        JB1_f1, JB1_p1 = jacobians_b1
        JB2_f2, JB2_p2 = jacobians_b2
        J_Bout_B1, J_Bout_B2 = jacobians_b_out
        J_out_f1 = torch.matmul(J_Bout_B1, JB1_f1)
        J_out_p1 = torch.matmul(J_Bout_B1, JB1_p1)
        J_out_f2 = torch.matmul(J_Bout_B2, JB2_f2)
        J_out_p2 = torch.matmul(J_Bout_B2, JB2_p2)

        return [J_out_f1, J_out_f2, J_out_p1, J_out_p2], error

    def dim(self) -> int:
        return self.frame1.dof()

    def _copy_impl(self, new_name: Optional[str] = None) -> "MovingFrameBetween":
        return MovingFrameBetween(
            self.frame1.copy(),
            self.frame2.copy(),
            self.pose1.copy(),
            self.pose2.copy(),
            self.measurement.copy(),
            self.weight.copy(),
            name=new_name,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, List, Sequence

from theseus.core import Objective, Variable


class VariableOrdering:
    def __init__(self, objective: Objective, default_order: bool = True):
        self.objective = objective
        self._var_order: List[Variable] = []
        self._var_name_to_index: Dict[str, int] = {}
        if default_order:
            self._compute_default_order(objective)

    def _compute_default_order(self, objective: Objective):
        assert not self._var_order and not self._var_name_to_index
        cur_idx = 0
        for variable_name, variable in objective.optim_vars.items():
            if variable_name in self._var_name_to_index:
                continue
            self._var_order.append(variable)
            self._var_name_to_index[variable_name] = cur_idx
            cur_idx += 1

    def index_of(self, key: str) -> int:
        return self._var_name_to_index[key]

    def __getitem__(self, index) -> Variable:
        return self._var_order[index]

    def __iter__(self):
        return iter(self._var_order)

    def append(self, var: Variable):
        if var in self._var_order:
            raise ValueError(
                f"Variable {var.name} has already been added to the order."
            )
        if var.name not in self.objective.optim_vars:
            raise ValueError(
                f"Variable {var.name} is not an optimization variable for the objective."
            )
        self._var_order.append(var)
        self._var_name_to_index[var.name] = len(self._var_order) - 1

    def remove(self, var: Variable):
        self._var_order.remove(var)
        del self._var_name_to_index[var.name]

    def extend(self, variables: Sequence[Variable]):
        for var in variables:
            self.append(var)

    @property
    def complete(self):
        return len(self._var_order) == self.objective.size_variables()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

__version__ = "0.1.0"

from .core import (
    CostFunction,
    CostWeight,
    DiagonalCostWeight,
    Objective,
    ScaleCostWeight,
    Variable,
    Vectorize,
    RobustLoss,
    AutoDiffCostFunction,
    RobustCostFunction,
    HuberLoss,
    WelschLoss,
)
from .geometry import (
    SE2,
    SE3,
    SO2,
    SO3,
    LieGroup,
    Manifold,
    Point2,
    Point3,
    Vector,
    adjoint,
    between,
    local,
    retract,
    compose,
    inverse,
    log_map,
    exp_map,
    LieGroupTensor,
    set_lie_tangent_enabled,
    enable_lie_tangent,
    no_lie_tangent,
    rand_vector,
    rand_point2,
    rand_point3,
    rand_so2,
    rand_so3,
    rand_se2,
    rand_se3,
    randn_vector,
    randn_point2,
    randn_point3,
    randn_so2,
    randn_so3,
    randn_se2,
    randn_se3,
)
from .optimizer import (
    DenseLinearization,
    SparseLinearization,
    VariableOrdering,
    ManifoldGaussian,
    local_gaussian,
    retract_gaussian,
)
from .optimizer.linear import (
    CholeskyDenseSolver,
    CholmodSparseSolver,
    DenseSolver,
    LinearOptimizer,
    LUCudaSparseSolver,
    LUDenseSolver,
)
from .optimizer.nonlinear import (
    GaussNewton,
    LevenbergMarquardt,
    NonlinearLeastSquares,
    NonlinearOptimizerParams,
    NonlinearOptimizerStatus,
    BackwardMode,
)
from .theseus_layer import TheseusLayer

import theseus.embodied as eb

# Aliases for some standard cost functions
Difference = eb.Local
Between = eb.Between
Local = eb.Local



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional, Tuple

import torch
from scipy import ndimage

from theseus.core import Variable
from theseus.utils import gather_from_rows_cols


class SignedDistanceField2D:
    # origin shape is:  batch_size x 2
    # sdf_data shape is: batch_size x field_height x field_width
    def __init__(
        self,
        origin: Variable,
        cell_size: Variable,
        sdf_data: Optional[Variable] = None,
        occupancy_map: Optional[Variable] = None,
    ):
        if occupancy_map is not None:
            if sdf_data is not None:
                raise ValueError(
                    "Only one of arguments sdf_data and occupancy_map should be provided."
                )
            sdf_data = self._compute_sdf_data_from_map(occupancy_map, cell_size)
        else:
            if sdf_data is None:
                raise ValueError(
                    "Either argument sdf_data or argument occupancy_map should be provided."
                )
        self.update_data(origin, sdf_data, cell_size)
        self._num_rows = sdf_data.shape[1]
        self._num_cols = sdf_data.shape[2]

    def _compute_sdf_data_from_map(
        self, occupancy_map_batch: Variable, cell_size: Variable
    ) -> Variable:
        # Code from https://github.com/gtrll/gpmp2/
        if occupancy_map_batch.ndim != 3:
            raise ValueError(
                "Argument occupancy_map to SignedDistanceField2D must be a batch of matrices."
            )
        num_maps = occupancy_map_batch.tensor.size(0)
        all_sdf_data = []

        for i in range(num_maps):
            occupancy_map = occupancy_map_batch[i]

            cur_map = occupancy_map > 0.75
            cur_map = cur_map.int()

            if torch.max(cur_map) == 0:
                map_x, map_y = occupancy_map.size(0), occupancy_map.size(1)
                max_map_size = 2 * cell_size[i].item() * max(map_x, map_y)
                sdf_data = (
                    torch.ones(occupancy_map.shape, dtype=cell_size.dtype)
                    * max_map_size
                )
            else:
                # inverse map
                inv_map = 1 - cur_map
                # get signed distance from map and inverse map
                # since bwdist(foo) = ndimage.distance_transform_edt(1-foo)
                map_dist = ndimage.distance_transform_edt(inv_map.numpy())
                inv_map_dist = ndimage.distance_transform_edt(cur_map.numpy())

                sdf_data = map_dist - inv_map_dist
                # metric
                sdf_data = torch.tensor(sdf_data, dtype=cell_size.dtype) * cell_size[i]

            all_sdf_data.append(sdf_data)

        sdf_data_var = Variable(torch.stack(all_sdf_data))

        return sdf_data_var

    def update_data(self, origin: Variable, sdf_data: Variable, cell_size: Variable):
        if sdf_data.ndim != 3:
            raise ValueError(
                "Argument sdf_data to SignedDistanceField2D must be a batch of matrices."
            )
        if not (origin.ndim == 2 or (origin.ndim == 3 and origin.shape[2] == 1)):
            raise ValueError(
                "Argument origin to SignedDistanceField2D must be a batch of 2-D tensors."
            )
        if not (
            cell_size.ndim == 1 or (cell_size.ndim == 2 and cell_size.shape[1] == 1)
        ):
            raise ValueError(
                "Argument cell_size must be a batch of 0-D or 1-D tensors."
            )
        if (
            origin.shape[0] != sdf_data.shape[0]
            or origin.shape[0] != cell_size.shape[0]
        ):
            raise ValueError("Incompatible batch size between input arguments.")
        self.origin = origin
        self.sdf_data = sdf_data
        self.cell_size = cell_size

    def convert_points_to_cell(
        self, points: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        origin = (
            self.origin.tensor.unsqueeze(-1)
            if self.origin.ndim == 2
            else self.origin.tensor
        )
        cell_size = (
            self.cell_size.tensor
            if self.cell_size.ndim == 2
            else self.cell_size.tensor.unsqueeze(-1)
        )
        px = points[:, 0]
        py = points[:, 1]

        out_of_bounds_idx = (
            (px < origin[:, 0])
            .logical_or(px > (origin[:, 0] + (self._num_cols - 1.0) * cell_size))
            .logical_or(py < origin[:, 1])
            .logical_or(py > (origin[:, 1] + (self._num_rows - 1.0) * cell_size))
        )

        col = (px - origin[:, 0]) / cell_size
        row = (py - origin[:, 1]) / cell_size
        return row, col, out_of_bounds_idx

    # Points shape must be (batch_size, 2, num_points)
    #   Computed distances for each point, shape (batch_size, num_points)
    #   Jacobian tensor, shape (batch_size, num_points, 2)
    def signed_distance(
        self, points: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        rows, cols, out_of_bounds_idx = self.convert_points_to_cell(points)

        lr = torch.floor(rows)
        lc = torch.floor(cols)
        hr = lr + 1.0
        hc = lc + 1.0
        lri = lr.long().clamp(0, self._num_rows - 1)
        lci = lc.long().clamp(0, self._num_cols - 1)
        hri = hr.long().clamp(0, self._num_rows - 1)
        hci = hc.long().clamp(0, self._num_cols - 1)

        def gather_sdf(r_, c_):
            return gather_from_rows_cols(self.sdf_data.tensor, r_, c_)

        # Compute the distance
        hrdiff = hr - rows
        hcdiff = hc - cols
        lrdiff = rows - lr
        lcdiff = cols - lc
        dist = (
            hrdiff * hcdiff * gather_sdf(lri, lci)
            + lrdiff * hcdiff * gather_sdf(hri, lci)
            + hrdiff * lcdiff * gather_sdf(lri, hci)
            + lrdiff * lcdiff * gather_sdf(hri, hci)
        )
        dist[out_of_bounds_idx] = 0

        # Compute the jacobians

        cell_size = (
            self.cell_size.tensor
            if self.cell_size.ndim == 2
            else self.cell_size.tensor.unsqueeze(-1)
        )

        jac1 = (
            hrdiff * (gather_sdf(lri, hci) - gather_sdf(lri, lci))
            + lrdiff * (gather_sdf(hri, hci) - gather_sdf(hri, lci))
        ) / cell_size
        jac2 = (
            hcdiff * (gather_sdf(hri, lci) - gather_sdf(lri, lci))
            + lcdiff * (gather_sdf(hri, hci) - gather_sdf(lri, hci))
        ) / cell_size
        jac1[out_of_bounds_idx] = 0
        jac2[out_of_bounds_idx] = 0
        return dist, torch.stack([jac1, jac2], dim=2)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Optional, Tuple

import torch

from theseus.core import CostFunction, CostWeight
from theseus.geometry import LieGroup, between


class Between(CostFunction):
    def __init__(
        self,
        v0: LieGroup,
        v1: LieGroup,
        measurement: LieGroup,
        cost_weight: CostWeight,
        name: Optional[str] = None,
    ):
        super().__init__(cost_weight, name=name)
        self.v0 = v0
        self.v1 = v1
        self.register_optim_vars(["v0", "v1"])
        self.measurement = measurement
        self.register_aux_vars(["measurement"])
        if not isinstance(v0, v1.__class__) or not isinstance(
            v0, measurement.__class__
        ):
            raise ValueError("Inconsistent types between variables and measurement.")

    def error(self) -> torch.Tensor:
        var_diff = between(self.v0, self.v1)
        return self.measurement.local(var_diff)

    def jacobians(self) -> Tuple[List[torch.Tensor], torch.Tensor]:
        Jlist: List[torch.Tensor] = []
        var_diff = between(self.v0, self.v1)
        log_jac: List[torch.Tensor] = []
        error = self.measurement.inverse().compose(var_diff).log_map(jacobians=log_jac)
        dlog = log_jac[0]
        Jlist.extend([-dlog @ var_diff.inverse().adjoint(), dlog])
        return Jlist, error

    def dim(self) -> int:
        return self.v0.dof()

    def _copy_impl(self, new_name: Optional[str] = None) -> "Between":
        return Between(
            self.v0.copy(),
            self.v1.copy(),
            self.measurement.copy(),
            self.weight.copy(),
            name=new_name,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .cost_function import AutoDiffCostFunction, CostFunction, ErrFnType
from .cost_weight import CostWeight, DiagonalCostWeight, ScaleCostWeight
from .objective import Objective
from .robust_cost_function import RobustCostFunction
from .robust_loss import HuberLoss, RobustLoss, WelschLoss
from .variable import Variable
from .vectorizer import Vectorize



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .collision import Collision2D
from .eff_obj_contact import EffectorObjectContactPlanar
from .signed_distance_field import SignedDistanceField2D



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from collections import defaultdict
from typing import Dict, Iterable, List, Optional, Set, Tuple, Type

import torch

from theseus.geometry.manifold import Manifold

from .cost_function import CostFunction
from .objective import Objective
from .variable import Variable

_CostFunctionSchema = Tuple[str, ...]


def _get_cost_function_schema(cost_function: CostFunction) -> _CostFunctionSchema:
    def _fullname(obj) -> str:
        return f"{obj.__module__}.{obj.__class__.__name__}"

    def _varinfo(var) -> str:
        return f"{_fullname(var)}{tuple(var.shape[1:])}"

    return (
        (_fullname(cost_function),)
        + tuple(_varinfo(v) for v in cost_function.optim_vars)
        + tuple(_varinfo(v) for v in cost_function.aux_vars)
        + (_fullname(cost_function.weight),)
        + tuple(_varinfo(v) for v in cost_function.weight.aux_vars)
    )


# This wrapper allows the weighted error and jacobians to be replaced by pre-computed
# values. The vectorization class will compute these in a batch for all cost function,
# then populate wrappers for each so they can be served to the linearization classes.
class _CostFunctionWrapper(CostFunction):
    def __init__(self, cost_fn: CostFunction, name: Optional[str] = None):
        if name is None:
            name = f"wrapper({cost_fn.name})"
        super().__init__(cost_fn.weight, name=name)
        self.cost_fn = cost_fn
        self.register_vars(cost_fn.optim_vars, is_optim_vars=True)
        self.register_vars(cost_fn.aux_vars, is_optim_vars=False)
        self._cached_error: Optional[torch.Tensor] = None
        self._cached_jacobians: Optional[List[torch.Tensor]] = None

    def error(self) -> torch.Tensor:
        return self.cost_fn.error()

    def jacobians(self) -> Tuple[List[torch.Tensor], torch.Tensor]:
        return self.cost_fn.jacobians()

    def weighted_error(self) -> torch.Tensor:
        if self._cached_error is None:
            return super().weighted_error()
        return self._cached_error

    def weighted_jacobians_error(
        self,
    ) -> Tuple[List[torch.Tensor], torch.Tensor]:
        if self._cached_jacobians is None:
            return super().weighted_jacobians_error()
        assert self._cached_error is not None
        return self._cached_jacobians, self._cached_error

    def dim(self) -> int:
        return self.cost_fn.dim()

    def _copy_impl(self, new_name: Optional[str] = None) -> "_CostFunctionWrapper":
        return _CostFunctionWrapper(self.cost_fn, name=new_name)

    def __str__(self) -> str:
        return f"Wrapper({str(_get_cost_function_schema(self.cost_fn))})"

    def __repr__(self) -> str:
        return str(self)


# This class replaces the Objective's iterator for one that takes advantage of
# cost function vectorization
class Vectorize:
    _SHARED_TOKEN = "__shared__"

    def __init__(self, objective: Objective, empty_cuda_cache: bool = False):
        # Each cost function is assigned a wrapper. The wrapper will hold the error
        # and jacobian after vectorization.
        self._cost_fn_wrappers: List[_CostFunctionWrapper] = []

        # This dict groups cost functions that can be vectorized together.
        self._schema_dict: Dict[
            _CostFunctionSchema, List[_CostFunctionWrapper]
        ] = defaultdict(list)

        # Create wrappers for all cost functions and also get their schemas
        for cost_fn in objective.cost_functions.values():
            wrapper = _CostFunctionWrapper(cost_fn)
            self._cost_fn_wrappers.append(wrapper)
            schema = _get_cost_function_schema(cost_fn)
            self._schema_dict[schema].append(wrapper)

        # Now create a vectorized cost function for each unique schema
        self._vectorized_cost_fns: Dict[_CostFunctionSchema, CostFunction] = {}
        for schema in self._schema_dict:
            base_cost_fn = self._schema_dict[schema][0].cost_fn
            vectorized_cost_fn = base_cost_fn.copy(keep_variable_names=False)
            vectorized_cost_fn.weight = base_cost_fn.weight.copy(
                keep_variable_names=False
            )
            self._vectorized_cost_fns[schema] = vectorized_cost_fn

        # Dict[_CostFunctionSchema, List[str]]
        self._var_names = self._get_var_names()

        # `self._vectorize()` will compute an error vector for each schema,
        # then populate the wrappers with their appropriate weighted error slice.
        # Replacing `obj._cost_functions_iterable` allows to recover these when
        # iterating the Objective.
        objective._enable_vectorization(
            self._cost_fn_wrappers,
            self._vectorize,
            self._to,
            self._vectorized_retract_optim_vars,
            self,
        )

        self._objective = objective
        self._empty_cuda_cache = empty_cuda_cache

    # Returns a dictionary which maps every schema to information about its shared
    # variables.
    # The information is in the form of a list of strings for all variables in the
    # schema, such that if the variable is shared the string has the name of the
    # variable and the prefix __SHARED__. Otherwise it returns "var_{idx}",
    # where idx is the index of the var in the schema order.
    def _get_var_names(self) -> Dict[_CostFunctionSchema, List[Optional[str]]]:
        info = {}
        for schema, cost_fn_wrappers in self._schema_dict.items():
            # schema is (_, N optim_vars, M aux_vars, _, P aux_vars)
            # so the total number of vars is len(schema) - 2
            # var_names_per_position holds all variable names associated to each
            # position in schema. Shared variables are those positions with only one
            # name
            var_names_per_position: List[Set[str]] = [
                set() for _ in range(len(schema) - 2)
            ]
            for wrapper in cost_fn_wrappers:
                cf = wrapper.cost_fn
                var_idx = 0
                for var_iterators in [cf.optim_vars, cf.aux_vars, cf.weight.aux_vars]:
                    for v in var_iterators:
                        var_names_per_position[var_idx].add(v.name)
                        var_idx += 1
                assert var_idx == len(schema) - 2

            def _get_name(var_idx, name_set_):
                if len(name_set_) == 1:
                    return f"{Vectorize._SHARED_TOKEN}{next(iter(name_set_))}"
                return f"var_{var_idx}"

            info[schema] = [
                _get_name(i, s) for i, s in enumerate(var_names_per_position)
            ]
        return info

    @staticmethod
    def _get_all_vars(cf: CostFunction) -> List[Variable]:
        return (
            list(cf.optim_vars)
            + list(cf.aux_vars)  # type:ignore
            + list(cf.weight.aux_vars)  # type:ignore
        )

    @staticmethod
    def _expand(tensor: torch.Tensor, size: int) -> torch.Tensor:
        return tensor.expand((size,) + ((-1,) * (len(tensor.shape) - 1)))

    # Populates names_to_tensors with a list with one element per vectorized variable,
    # and which holds the concatenated tensors of its corresponding variables in all
    # cost functions in the list (which are assumed to have the same schema).
    # Inputs:
    #   - vars_names; The names for all variables in the cost_fns' schema. Shared
    #       variables must be prefixed with `_SHARED_TOKEN`.`
    #   - batch_size: the Objectve's batch size. Whenever a cost function's var tensor
    #       is batch_size 1, it expands it to this value, unless the variable is
    #       shared by all cost functions.
    #   - names_to_tensors: A dictionary mapping variable names to the list of tensors.
    #       It will be modified in place.
    @staticmethod
    def _update_all_cost_fns_var_tensors(
        cf_wrappers: List[_CostFunctionWrapper],
        var_names: List[str],
        objective_batch_size: int,
        names_to_tensors: Dict[str, List[torch.Tensor]],
    ):
        # Get all the tensors from individual variables
        for wrapper in cf_wrappers:
            cost_fn = wrapper.cost_fn
            cost_fn_vars = Vectorize._get_all_vars(cost_fn)
            # some shared vars may appear in more than one position in the list
            # For example GPMotionModel and GPCostWeight both have dt in their
            # list of vars. With this set we can skip repetitions
            seen_vars = set()
            for var_idx, var in enumerate(cost_fn_vars):
                name = var_names[var_idx]
                if name in seen_vars:
                    continue

                # If the variable is shared and batch_size == 1, only need a tensor for
                # one of the cost functions and we can just extend later to complete
                # the vectorized batch, so here we can do `continue`
                var_batch_size = var.tensor.shape[0]
                if (
                    name in names_to_tensors
                    and Vectorize._SHARED_TOKEN in name
                    and var_batch_size == 1
                ):
                    continue

                # Otherwise, we need to append to the list of tensors. since
                # we cannot extend to a full batch w/o copying.

                # If not a shared variable, expand to batch size if needed, because
                # those we will copy, no matter what.
                # For shared variables, just append, we will handle the expansion
                # when updating the vectorized variable containers.
                tensor = (
                    var.tensor
                    if (var_batch_size > 1 or Vectorize._SHARED_TOKEN in name)
                    else Vectorize._expand(var.tensor, objective_batch_size)
                )
                names_to_tensors[name].append(tensor)
                seen_vars.add(name)

    # Goes through the list of vectorized variables and updates them with the
    # concatenation of all tensors in their corresponding entry in
    # `names_to_tensors`. Shared variables are expanded to shape
    # batch_size * num_cost_fns
    @staticmethod
    def _update_vectorized_vars(
        all_vectorized_vars: List[Variable],
        names_to_tensors: Dict[str, List[torch.Tensor]],
        var_names: List[str],
        batch_size: int,
        num_cost_fns: int,
    ):
        for var_idx, var in enumerate(all_vectorized_vars):
            name = var_names[var_idx]

            if num_cost_fns == 1:
                var.update(names_to_tensors[name][0])
                continue

            all_var_tensors = names_to_tensors[name]
            if Vectorize._SHARED_TOKEN in name and all_var_tensors[0].shape[0] == 1:
                # In this case this is a shared variable, so all_var_tensors[i] is
                # the same for any value of i. So, we can just expand to the full
                # vectorized size. Sadly, this doesn't work if batch_size > 1
                var_tensor = Vectorize._expand(
                    all_var_tensors[0], batch_size * num_cost_fns
                )
            else:
                var_tensor = torch.cat(all_var_tensors, dim=0)
            var.update(var_tensor)

    # Computes the error of the vectorized cost function and distributes the error
    # to the cost function wrappers. The list of wrappers must correspond to the
    # same schema from which the vectorized cost function was obtained.
    @staticmethod
    def _compute_error_and_replace_wrapper_caches(
        vectorized_cost_fn: CostFunction,
        cost_fn_wrappers: List[_CostFunctionWrapper],
        batch_size: int,
    ):
        v_jac, v_err = vectorized_cost_fn.weighted_jacobians_error()
        start_idx = 0
        for wrapper in cost_fn_wrappers:
            assert wrapper._cached_error is None
            assert wrapper._cached_jacobians is None
            v_slice = slice(start_idx, start_idx + batch_size)
            wrapper._cached_error = v_err[v_slice]
            wrapper._cached_jacobians = [jac[v_slice] for jac in v_jac]
            start_idx += batch_size

    def _clear_wrapper_caches(self):
        for cost_fn_wrappers in self._schema_dict.values():
            for cf in cost_fn_wrappers:
                cf._cached_error = None
                cf._cached_jacobians = None
        if self._empty_cuda_cache:
            torch.cuda.empty_cache()

    # This could be a static method, but writing like this makes some unit tests
    # easier
    def _handle_singleton_wrapper(
        self, _: _CostFunctionSchema, cost_fn_wrappers: List[_CostFunctionWrapper]
    ):
        wrapper = cost_fn_wrappers[0]
        (
            wrapper._cached_jacobians,
            wrapper._cached_error,
        ) = wrapper.cost_fn.weighted_jacobians_error()

    def _handle_schema_vectorization(
        self, schema: _CostFunctionSchema, cost_fn_wrappers: List[_CostFunctionWrapper]
    ):
        var_names = self._var_names[schema]
        vectorized_cost_fn = self._vectorized_cost_fns[schema]
        all_vectorized_vars = Vectorize._get_all_vars(vectorized_cost_fn)
        assert len(all_vectorized_vars) == len(var_names)
        names_to_tensors: Dict[str, List[torch.Tensor]] = defaultdict(list)
        batch_size = self._objective.batch_size

        Vectorize._update_all_cost_fns_var_tensors(
            cost_fn_wrappers, var_names, batch_size, names_to_tensors
        )
        Vectorize._update_vectorized_vars(
            all_vectorized_vars,
            names_to_tensors,
            var_names,
            batch_size,
            len(cost_fn_wrappers),
        )
        Vectorize._compute_error_and_replace_wrapper_caches(
            vectorized_cost_fn, cost_fn_wrappers, batch_size
        )

    def _vectorize(self):
        self._clear_wrapper_caches()
        for schema, cost_fn_wrappers in self._schema_dict.items():
            if len(cost_fn_wrappers) == 1:
                self._handle_singleton_wrapper(schema, cost_fn_wrappers)
            else:
                self._handle_schema_vectorization(schema, cost_fn_wrappers)

    @staticmethod
    def _vectorized_retract_optim_vars(
        delta: torch.Tensor,
        ordering: Iterable[Manifold],
        ignore_mask: Optional[torch.Tensor] = None,
        force_update: bool = False,
    ):
        # Each (variable-type, dof) gets mapped to a tuple with:
        #   - the variable that will hold the vectorized tensor
        #   - all the variables of that type that will be vectorized together
        #   - the delta slices that go into the batch
        var_info: Dict[
            Tuple[Type[Manifold], int],
            Tuple[Manifold, List[Manifold], List[torch.Tensor]],
        ] = {}

        start_idx = 0
        batch_size = -1
        # Create var info by looping variables in the given order
        # All variables of the same type get grouped together, and their deltas saved
        for var in ordering:
            if batch_size == -1:
                batch_size = var.shape[0]
            else:
                assert batch_size == var.shape[0]

            delta_slice = delta[:, start_idx : start_idx + var.dof()]
            start_idx += var.dof()
            var_type = (var.__class__, var.dof())
            if var_type not in var_info:
                var_info[var_type] = (var.copy(), [], [])
            var_info[var_type][1].append(var)
            var_info[var_type][2].append(delta_slice)

        for _, (vectorized_var, var_list, delta_list) in var_info.items():
            n_vars, dof = len(var_list), vectorized_var.dof()

            # Get the vectorized tensor that holds all the current variable tensors.
            # The resulting shape is (N * b, M), b is batch size, N is the number of
            # variables in the group, and M is the data shape for this class
            vectorized_tensor = torch.cat([v.tensor for v in var_list], dim=0)
            assert (
                vectorized_tensor.shape
                == (n_vars * batch_size,) + vectorized_tensor.shape[1:]
            )
            # delta_list is a list of N tensors of shape (b, d), where
            # d is var.dof(). After vectorization, we get tensor of shape (N * b, d)
            vectorized_delta = torch.cat(delta_list, dim=0)
            assert vectorized_delta.shape == (n_vars * batch_size, dof)

            # Retract the vectorized tensor, then redistribute to the variables
            vectorized_var.update(vectorized_tensor)
            new_var = vectorized_var.retract(vectorized_delta)
            start_idx = 0
            for var in var_list:
                retracted_tensor_slice = new_var[start_idx : start_idx + batch_size, :]
                if ignore_mask is None or force_update:
                    var.update(retracted_tensor_slice)
                else:
                    var.update(retracted_tensor_slice, batch_ignore_mask=ignore_mask)
                start_idx += batch_size

    # Applies to() with given args to all vectorized cost functions in the objective
    def _to(self, *args, **kwargs):
        for cost_function in self._vectorized_cost_fns.values():
            cost_function.to(*args, **kwargs)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Optional, Tuple, Union, cast

import torch

from theseus.core import CostFunction, CostWeight, Variable
from theseus.embodied.kinematics import IdentityModel
from theseus.geometry import SE2, Point2

from .signed_distance_field import SignedDistanceField2D


class EffectorObjectContactPlanar(CostFunction):
    def __init__(
        self,
        obj: SE2,
        eff: SE2,
        sdf_origin: Variable,
        sdf_data: Variable,
        sdf_cell_size: Variable,
        eff_radius: Union[float, Variable, torch.Tensor],
        cost_weight: CostWeight,
        name: Optional[str] = None,
        use_huber_loss: bool = False,
    ):
        super().__init__(cost_weight, name=name)
        self.obj = obj
        self.eff = eff
        self.sdf_origin = sdf_origin
        self.sdf_data = sdf_data
        self.sdf_cell_size = sdf_cell_size
        if not isinstance(eff_radius, Variable):
            if not isinstance(eff_radius, torch.Tensor):
                eff_radius = torch.tensor(eff_radius)
            self.eff_radius = Variable(eff_radius)
        else:
            self.eff_radius = eff_radius
        if self.eff_radius.tensor.squeeze().ndim > 1:
            raise ValueError("eff_radius must be a 0-D or 1-D tensor.")
        self.eff_radius.tensor = self.eff_radius.tensor.view(-1, 1)
        self.register_optim_vars(["obj", "eff"])
        self.register_aux_vars(
            ["sdf_origin", "sdf_data", "sdf_cell_size", "eff_radius"]
        )
        self.robot = IdentityModel()
        self.sdf = SignedDistanceField2D(sdf_origin, sdf_cell_size, sdf_data)
        self._use_huber = use_huber_loss

        if use_huber_loss:
            raise NotImplementedError(
                "Jacobians for huber loss are not yet implemented."
            )

    def _compute_distances_and_jacobians(
        self,
    ) -> Tuple[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        J_transf: List[torch.Tensor] = []
        J_xy: List[torch.Tensor] = []
        eff__obj = self.obj.transform_to(
            self.eff.xy(jacobians=J_xy), jacobians=J_transf
        )
        J_transf_obj = J_transf[0]
        J_transf_eff = J_transf[1].matmul(J_xy[0])
        robot_state = cast(Point2, self.robot.forward_kinematics(eff__obj)["state"])
        dist, J_dist = self.sdf.signed_distance(robot_state.tensor.view(-1, 2, 1))
        J_out = (J_dist.matmul(J_transf_obj), J_dist.matmul(J_transf_eff))
        return dist, J_out

    def _error_from_distances(self, distances: torch.Tensor):
        if self._use_huber:
            eff_rad = self.eff_radius.tensor
            err = distances.clone()
            # linear (two-sided, otherwise this would be 0)
            gt_r_idx = distances >= 2 * eff_rad
            err[gt_r_idx] = distances[gt_r_idx] - 0.5 * eff_rad
            # quadratic
            lt_r_pos_idx = (distances < 2 * eff_rad).logical_or(distances > 0)
            err[lt_r_pos_idx] = (
                0.5 * (distances[lt_r_pos_idx] - eff_rad).square() / eff_rad
            )
            # linear
            neg_idx = distances < 0
            err[neg_idx] = -distances[neg_idx] + 0.5 * eff_rad
            return err
        else:
            return (distances - self.eff_radius.tensor).abs()

    def error(self) -> torch.Tensor:
        distances, _ = self._compute_distances_and_jacobians()
        return self._error_from_distances(distances)

    def jacobians(self) -> Tuple[List[torch.Tensor], torch.Tensor]:
        distances, jacobians = self._compute_distances_and_jacobians()
        error = self._error_from_distances(distances)
        if self._use_huber:
            raise NotImplementedError(
                "Jacobians for huber loss are not yet implemented."
            )
        else:
            lt_idx = distances < self.eff_radius.tensor
            jacobians[0][lt_idx] *= -1.0
            jacobians[1][lt_idx] *= -1.0
        return [jacobians[0], jacobians[1]], error

    def _copy_impl(
        self, new_name: Optional[str] = None
    ) -> "EffectorObjectContactPlanar":
        return EffectorObjectContactPlanar(
            self.obj.copy(),
            self.eff.copy(),
            self.sdf_origin.copy(),
            self.sdf_data.copy(),
            self.sdf_cell_size.copy(),
            self.eff_radius.copy(),
            self.weight.copy(),
            name=new_name,
        )

    def dim(self) -> int:
        return 1

    # This is needed so that the SDF container also updates with the new aux var
    def set_aux_var_at(self, index: int, variable: Variable):
        super().set_aux_var_at(index, variable)
        self.sdf.update_data(self.sdf_origin, self.sdf_data, self.sdf_cell_size)



import collections
import pathlib
from typing import Dict, List, Optional, Tuple, cast

import numpy as np
import omegaconf
import torch
import torch.nn as nn

import theseus as th

from .misc import TactilePushingDataset


class TactileMeasModel(nn.Module):
    def __init__(self, input_size: int, output_size: int):
        super().__init__()

        self.fc1 = nn.Linear(input_size, output_size)

    def forward(self, x1: torch.Tensor, x2: torch.Tensor, k: torch.Tensor):
        x = torch.cat([x1, x2], dim=1)

        k1_ = k.unsqueeze(1)  # b x 1 x cl
        x1_ = x.unsqueeze(-1)  # b x dim x 1
        x = torch.mul(x1_, k1_)  # b x dim x cl

        x = x.view(x.shape[0], -1)
        x = self.fc1(x)

        return x


def init_tactile_model_from_file(model: nn.Module, filename: pathlib.Path):

    model_saved = torch.jit.load(filename)
    state_dict_saved = model_saved.state_dict()
    state_dict_new = collections.OrderedDict()
    state_dict_new["fc1.weight"] = state_dict_saved["model.fc1.weight"]
    state_dict_new["fc1.bias"] = state_dict_saved["model.fc1.bias"]

    model.load_state_dict(state_dict_new)

    return model


# Set some parameters for the cost weights
class TactileWeightModel(nn.Module):
    def __init__(
        self, device: torch.device, dim: int = 3, wt_init: Optional[torch.Tensor] = None
    ):
        super().__init__()

        wt_init_ = torch.rand(1, dim)
        if wt_init is not None:
            wt_init_ = wt_init
        self.param = nn.Parameter(wt_init_.to(device))
        self.to(device)

    def forward(self):
        return self.param.clone()


def create_tactile_models(
    model_type: str,
    device: torch.device,
    measurements_model_path: Optional[pathlib.Path] = None,
) -> Tuple[
    Optional[TactileMeasModel],
    TactileWeightModel,
    TactileWeightModel,
    List[nn.Parameter],
]:
    if model_type == "weights_only":
        qsp_model = TactileWeightModel(
            device, wt_init=torch.tensor([[50.0, 50.0, 50.0]])
        )
        mf_between_model = TactileWeightModel(
            device, wt_init=torch.tensor([[0.0, 0.0, 10.0]])
        )
        measurements_model = None

        learnable_params = list(qsp_model.parameters()) + list(
            mf_between_model.parameters()
        )
    elif model_type == "weights_and_measurement_nn":
        qsp_model = TactileWeightModel(device, wt_init=torch.tensor([[5.0, 5.0, 5.0]]))
        mf_between_model = TactileWeightModel(
            device, wt_init=torch.tensor([[0.0, 0.0, 5.0]])
        )
        measurements_model = TactileMeasModel(2 * 2 * 4, 4)
        if measurements_model_path is not None:
            measurements_model = init_tactile_model_from_file(
                model=measurements_model,
                filename=measurements_model_path,
            )
        measurements_model.to(device)

        learnable_params = (
            list(measurements_model.parameters())
            + list(qsp_model.parameters())
            + list(mf_between_model.parameters())
        )
    else:
        print("Learning mode {cfg.train.mode} not found")

    return (
        measurements_model,
        qsp_model,
        mf_between_model,
        learnable_params,
    )


# ----------------------------------------------------------------------------------- #
# ------------------------------- Theseus Model Interface --------------------------- #
# ----------------------------------------------------------------------------------- #


def get_tactile_nn_measurements_inputs(
    batch: Dict[str, torch.Tensor],
    device: torch.device,
    class_label: int,
    num_classes: int,
    min_win_mf: int,
    max_win_mf: int,
    step_win_mf: int,
    time_steps: int,
    model: Optional[nn.Module] = None,
):
    inputs = {}

    if model is not None:
        images_feat_meas = batch["img_feats"].to(device)
        class_label_vec = (
            nn.functional.one_hot(torch.tensor(class_label), torch.tensor(num_classes))
            .view(1, -1)
            .to(device)
        )

        meas_model_input_1_list: List[torch.Tensor] = []
        meas_model_input_2_list: List[torch.Tensor] = []
        for i in range(min_win_mf, time_steps):
            for offset in range(min_win_mf, np.minimum(i, max_win_mf), step_win_mf):
                meas_model_input_1_list.append(images_feat_meas[:, i - offset, :])
                meas_model_input_2_list.append(images_feat_meas[:, i, :])

        meas_model_input_1 = torch.cat(meas_model_input_1_list, dim=0)
        meas_model_input_2 = torch.cat(meas_model_input_2_list, dim=0)
        num_measurements = meas_model_input_1.shape[0]
        model_measurements = model.forward(
            meas_model_input_1, meas_model_input_2, class_label_vec
        ).reshape(
            -1, num_measurements, 4
        )  # data format (x, y, cos, sin)
    else:  # use oracle model
        eff_poses = batch["eff_poses"]
        obj_poses = batch["obj_poses"]
        model_measurements = []
        for i in range(min_win_mf, time_steps):
            for offset in range(min_win_mf, np.minimum(i, max_win_mf), step_win_mf):
                eff_pose_1 = th.SE2(x_y_theta=eff_poses[:, i - offset])
                obj_pose_1 = th.SE2(x_y_theta=obj_poses[:, i - offset])
                eff_pose_1__obj = obj_pose_1.between(eff_pose_1)

                eff_pose_2 = th.SE2(x_y_theta=eff_poses[:, i])
                obj_pose_2 = th.SE2(x_y_theta=obj_poses[:, i])
                eff_pose_2__obj = obj_pose_2.between(eff_pose_2)

                meas_pose_rel = cast(th.SE2, eff_pose_1__obj.between(eff_pose_2__obj))
                model_measurements.append(
                    torch.cat(
                        (
                            meas_pose_rel.xy(),
                            meas_pose_rel.theta().cos(),
                            meas_pose_rel.theta().sin(),
                        ),
                        dim=1,
                    )
                )  # data format (x, y, cos, sin)

        num_measurements = len(model_measurements)
        model_measurements = torch.cat(model_measurements, dim=0).reshape(
            -1, num_measurements, 4
        )
        model_measurements = model_measurements.to(device)

    # set MovingFrameBetween measurements from the NN output
    nn_meas_idx = 0
    for i in range(min_win_mf, time_steps):
        for offset in range(min_win_mf, np.minimum(i, max_win_mf), step_win_mf):
            meas_xycs_ = torch.stack(
                [
                    model_measurements[:, nn_meas_idx, 0],
                    model_measurements[:, nn_meas_idx, 1],
                    model_measurements[:, nn_meas_idx, 2],
                    model_measurements[:, nn_meas_idx, 3],
                ],
                dim=1,
            )
            inputs[f"nn_measurement_{i-offset}_{i}"] = meas_xycs_
            nn_meas_idx = nn_meas_idx + 1

    return inputs


def get_tactile_motion_capture_inputs(
    batch: Dict[str, torch.Tensor], device: torch.device, time_steps: int
):
    inputs = {}
    captures = batch["eff_poses"].to(device)
    for step in range(time_steps):
        capture = captures[:, step, :]
        cature_xycs = torch.stack(
            [capture[:, 0], capture[:, 1], capture[:, 2].cos(), capture[:, 2].sin()],
            dim=1,
        )
        inputs[f"motion_capture_{step}"] = cature_xycs
    return inputs


def get_tactile_cost_weight_inputs(qsp_model, mf_between_model):
    return {"qsp_weight": qsp_model(), "mf_between_weight": mf_between_model()}


def get_tactile_initial_optim_vars(
    batch: Dict[str, torch.Tensor],
    device: torch.device,
    time_steps: int,
):
    inputs = {}
    eff_captures = batch["eff_poses"].to(device)
    obj_captures = batch["obj_poses"].to(device)
    for step in range(time_steps):
        inputs[f"obj_pose_{step}"] = th.SE2(x_y_theta=obj_captures[:, 0].clone()).tensor
        inputs[f"eff_pose_{step}"] = th.SE2(x_y_theta=eff_captures[:, 0].clone()).tensor

    return inputs


def update_tactile_pushing_inputs(
    dataset: TactilePushingDataset,
    batch: Dict[str, torch.Tensor],
    measurements_model: nn.Module,
    qsp_model: nn.Module,
    mf_between_model: nn.Module,
    device: torch.device,
    cfg: omegaconf.DictConfig,
    theseus_inputs: Dict[str, torch.Tensor],
):
    time_steps = dataset.time_steps
    theseus_inputs["sdf_data"] = (dataset.sdf_data_tensor).to(device)
    theseus_inputs["sdf_cell_size"] = dataset.sdf_cell_size.to(device)
    theseus_inputs["sdf_origin"] = dataset.sdf_origin.to(device)

    theseus_inputs.update(
        get_tactile_nn_measurements_inputs(
            batch=batch,
            device=device,
            class_label=cfg.class_label,
            num_classes=cfg.num_classes,
            min_win_mf=cfg.tactile_cost.min_win_mf,
            max_win_mf=cfg.tactile_cost.max_win_mf,
            step_win_mf=cfg.tactile_cost.step_win_mf,
            time_steps=time_steps,
            model=measurements_model,
        )
    )
    theseus_inputs.update(get_tactile_motion_capture_inputs(batch, device, time_steps))
    theseus_inputs.update(get_tactile_cost_weight_inputs(qsp_model, mf_between_model))
    theseus_inputs.update(get_tactile_initial_optim_vars(batch, device, time_steps))


def get_tactile_poses_from_values(
    values: Dict[str, torch.Tensor], time_steps
) -> Tuple[torch.Tensor, torch.Tensor]:
    batch_size = values["obj_pose_0"].shape[0]
    device = values["obj_pose_0"].device

    obj_poses = torch.empty(batch_size, time_steps, 3, device=device)
    eff_poses = torch.empty(batch_size, time_steps, 3, device=device)

    for key in ["obj", "eff"]:
        ret_tensor = obj_poses if key == "obj" else eff_poses
        for t_ in range(time_steps):
            ret_tensor[:, t_, :2] = values[f"{key}_pose_{t_}"][:, 0:2]
            ret_tensor[:, t_, 2] = torch.atan2(
                values[f"{key}_pose_{t_}"][:, 3], values[f"{key}_pose_{t_}"][:, 2]
            )
    return obj_poses, eff_poses



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import pytest  # noqa: F401
import torch

import theseus as th
from theseus.core.vectorizer import _CostFunctionWrapper


def test_costs_vars_and_err_before_vectorization():
    for _ in range(20):
        objective = th.Objective()
        batch_size = torch.randint(low=1, high=10, size=(1,)).item()
        v1 = th.Vector(tensor=torch.randn(batch_size, 1), name="v1")
        v2 = th.Vector(tensor=torch.randn(batch_size, 1), name="v2")
        odummy = th.Vector(1, name="odummy")
        t1 = th.Vector(tensor=torch.zeros(1, 1), name="t1")
        adummy = th.Variable(tensor=torch.zeros(1, 1), name="adummy")
        cw1 = th.ScaleCostWeight(th.Variable(torch.zeros(1, 1), name="w1"))
        cw2 = th.ScaleCostWeight(th.Variable(torch.zeros(1, 1), name="w2"))
        cf1 = th.Difference(v1, t1, cw1)

        # Also test with autodiff cost
        def err_fn(optim_vars, aux_vars):
            return optim_vars[0].tensor - aux_vars[0].tensor

        cf2 = th.AutoDiffCostFunction([v2, odummy], err_fn, 1, cw2, [t1, adummy])

        # Chech that vectorizer's has the correct number of wrappers
        objective.add(cf1)
        objective.add(cf2)
        th.Vectorize(objective)

        # Update weights after creating vectorizer to see if data is picked up correctly
        w1 = torch.randn(1, 1)  # also check that broadcasting works
        w2 = torch.randn(batch_size, 1)

        # disable for this test since we are not checking the result
        objective._vectorization_needs_update = lambda: False
        objective.update({"w1": w1, "w2": w2})

        def _check_attr(cf, var):
            return hasattr(cf, var.name) and getattr(cf, var.name) is var

        # Check that the vectorizer's cost functions have the right variables and error
        saw_cf1 = False
        saw_cf2 = False
        for cf in objective._get_iterator():
            assert isinstance(cf, _CostFunctionWrapper)
            optim_vars = [v for v in cf.optim_vars]
            aux_vars = [v for v in cf.aux_vars]
            assert t1 in aux_vars
            assert _check_attr(cf, t1)
            w_err = cf.weighted_error()
            if cf.cost_fn is cf1:
                assert v1 in optim_vars
                assert w_err.allclose((v1.tensor - t1.tensor) * w1)
                assert _check_attr(cf, v1)
                saw_cf1 = True
            elif cf.cost_fn is cf2:
                assert v2 in optim_vars and odummy in optim_vars
                assert adummy in aux_vars
                assert _check_attr(cf, v2) and _check_attr(cf, odummy)
                assert w_err.allclose((v2.tensor - t1.tensor) * w2)
                saw_cf2 = True
            else:
                assert False
        assert saw_cf1 and saw_cf2


def test_correct_schemas_and_shared_vars():
    v1 = th.Vector(1)
    v2 = th.Vector(1)
    tv = th.Vector(1)
    w1 = th.ScaleCostWeight(1.0)
    mv = th.Vector(1)

    v3 = th.Vector(3)
    v4 = th.Vector(3)

    s1 = th.SE2()
    s2 = th.SE2()
    ts = th.SE2()

    objective = th.Objective()
    # these two can be grouped
    cf1 = th.Difference(v1, tv, w1)
    cf2 = th.Difference(v2, tv, w1)
    objective.add(cf1)
    objective.add(cf2)

    # this one uses the same weight and v1, v2, but cannot be grouped
    cf3 = th.Between(v1, v2, mv, w1)
    objective.add(cf3)

    # this one is the same cost function type, var type, and weight but different
    # dimension, so cannot be grouped either
    cf4 = th.Difference(v3, v4, w1)
    objective.add(cf4)

    # Now add another group with a different data-type (no-shared weight)
    w2 = th.ScaleCostWeight(1.0)
    w3 = th.ScaleCostWeight(2.0)
    cf5 = th.Difference(s1, ts, w2)
    cf6 = th.Difference(s2, ts, w3)
    objective.add(cf5)
    objective.add(cf6)

    # Not grouped with anything cf1 and cf2 because weight type is different
    w7 = th.DiagonalCostWeight([1.0])
    cf7 = th.Difference(v1, tv, w7)
    objective.add(cf7)

    vectorization = th.Vectorize(objective)

    assert len(vectorization._schema_dict) == 5
    seen_cnt = [0] * 7
    for schema, cost_fn_wrappers in vectorization._schema_dict.items():
        cost_fns = [w.cost_fn for w in cost_fn_wrappers]
        var_names = vectorization._var_names[schema]
        if cf1 in cost_fns:
            assert len(cost_fns) == 2
            assert cf2 in cost_fns
            seen_cnt[0] += 1
            seen_cnt[1] += 1
            assert f"{th.Vectorize._SHARED_TOKEN}{w1.scale.name}" in var_names
            assert f"{th.Vectorize._SHARED_TOKEN}{tv.name}" in var_names
        if cf3 in cost_fns:
            assert len(cost_fns) == 1
            seen_cnt[2] += 1
        if cf4 in cost_fns:
            assert len(cost_fns) == 1
            seen_cnt[3] += 1
        if cf5 in cost_fns:
            assert len(cost_fns) == 2
            assert cf6 in cost_fns
            seen_cnt[4] += 1
            seen_cnt[5] += 1
            assert f"{th.Vectorize._SHARED_TOKEN}{w2.scale.name}" not in var_names
            assert f"{th.Vectorize._SHARED_TOKEN}{w3.scale.name}" not in var_names
            assert f"{th.Vectorize._SHARED_TOKEN}{ts.name}" in var_names
        if cf7 in cost_fns:
            assert len(cost_fns) == 1
            seen_cnt[6] += 1
    assert seen_cnt == [1] * 7


def test_vectorized_error():
    rng = np.random.default_rng(0)
    generator = torch.Generator()
    generator.manual_seed(0)
    for _ in range(20):
        dim = rng.choice([1, 2])
        objective = th.Objective()
        batch_size = rng.choice(range(1, 11))

        vectors = [
            th.Vector(
                tensor=torch.randn(batch_size, dim, generator=generator), name=f"v{i}"
            )
            for i in range(rng.choice([1, 10]))
        ]
        target = th.Vector(dim, name="target")
        w = th.ScaleCostWeight(torch.randn(1, generator=generator))
        for v in vectors:
            objective.add(th.Difference(v, target, w))

        se3s = [
            th.SE3(
                tensor=th.SE3.rand(batch_size, generator=generator).tensor,
                strict=False,
            )
            for i in range(rng.choice([1, 10]))
        ]
        s_target = th.SE3.rand(1, generator=generator)
        ws = th.DiagonalCostWeight(torch.randn(6, generator=generator))
        for s in se3s:
            objective.add(th.Difference(s, s_target, ws))

        vectorization = th.Vectorize(objective)
        objective.update_vectorization_if_needed()

        assert objective._cost_functions_iterable is vectorization._cost_fn_wrappers
        for w in vectorization._cost_fn_wrappers:
            for cost_fn in objective.cost_functions.values():
                if cost_fn is w.cost_fn:
                    w_jac, w_err = cost_fn.weighted_jacobians_error()
                    assert w._cached_error.allclose(w_err)
                    for jac, exp_jac in zip(w._cached_jacobians, w_jac):
                        assert jac.allclose(exp_jac, atol=1e-6)

        squared_error = torch.cat(
            [cf.weighted_error() for cf in objective.cost_functions.values()], dim=1
        )
        assert squared_error.allclose(objective.error())


def test_vectorized_retract():
    rng = np.random.default_rng(0)
    generator = torch.Generator()
    generator.manual_seed(0)
    for _ in range(100):
        variables = []
        deltas = []
        batch_size = rng.choice(range(1, 11))
        n_vars = rng.choice([1, 10, 100])
        for _ in range(n_vars):
            var_type: th.LieGroup = rng.choice(
                [th.Vector, th.SE2, th.SE3, th.SO2, th.SO3]
            )
            if var_type == th.Vector:
                dof = rng.integers(1, 10)
                var = th.Vector.rand(batch_size, dof, generator=generator)
            else:
                var = var_type.rand(batch_size, generator=generator)
            deltas.append(torch.randn((batch_size, var.dof()), generator=generator))
            variables.append(var)
        variables_vectorized = [v.copy() for v in variables]
        delta = torch.cat(deltas, dim=1)

        ignore_mask = torch.rand(batch_size, generator=generator) > 0.5
        force_update = rng.random() > 0.5
        th.Objective._retract_base(
            delta, variables, ignore_mask=ignore_mask, force_update=force_update
        )
        th.Vectorize._vectorized_retract_optim_vars(
            delta,
            variables_vectorized,
            ignore_mask=ignore_mask,
            force_update=force_update,
        )

        for v1, v2 in zip(variables, variables_vectorized):
            assert v1.tensor.allclose(v2.tensor)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import pytest  # noqa: F401
import torch

import theseus as th
from theseus.constants import EPS

from .common import (
    check_jacobian_for_local,
    check_projection_for_exp_map,
    check_projection_for_log_map,
)


def test_xy_point2():
    for _ in range(100):
        for batch_size in [1, 10, 100]:
            point = th.Point2(tensor=torch.randn(batch_size, 2))
            assert point.x().allclose(point.tensor[:, 0])
            assert point.y().allclose(point.tensor[:, 1])


def test_xyz_point3():
    for _ in range(100):
        for batch_size in [1, 10, 100]:
            point = th.Point3(tensor=torch.randn(batch_size, 3))
            assert point.x().allclose(point.tensor[:, 0])
            assert point.y().allclose(point.tensor[:, 1])
            assert point.z().allclose(point.tensor[:, 2])


def test_point_operations_return_correct_type():
    for point_cls in [th.Point2, th.Point3]:
        p1 = point_cls()
        p2 = point_cls()

        assert isinstance(p1 + p2, point_cls)
        assert isinstance(p1 - p2, point_cls)
        assert isinstance(p1 * p2, point_cls)
        assert isinstance(p1 / p2, point_cls)
        assert isinstance(p1.abs(), point_cls)
        assert isinstance(-p1, point_cls)
        assert isinstance(p1.compose(p2), point_cls)
        assert isinstance(p1.retract(p2.tensor), point_cls)

        # for these, test result also since this method was overridden
        p1_copy = p1.copy()
        assert isinstance(p1_copy, point_cls)
        assert p1_copy.allclose(p1)
        exp_map = point_cls.exp_map(p2.tensor)
        assert isinstance(exp_map, point_cls)
        assert exp_map.allclose(p2)


def test_operations_mypy_cast():
    # mypy is optional install, only needed for library contributors
    try:
        import mypy.api
    except ModuleNotFoundError:
        return
    result = mypy.api.run(["theseus/geometry/tests/point_types_mypy_check.py"])
    assert result[2] == 0


def test_exp_map():
    rng = torch.Generator()
    rng.manual_seed(0)

    for batch_size in [1, 20, 100]:
        tangent_vector = torch.rand(batch_size, 2, generator=rng).double() - 0.5
        ret = th.Point2.exp_map(tangent_vector)

        assert torch.allclose(ret.tensor, tangent_vector, atol=EPS)
        check_projection_for_exp_map(
            tangent_vector, Group=th.Point2, is_projected=False
        )

    for batch_size in [1, 20, 100]:
        tangent_vector = torch.rand(batch_size, 3, generator=rng).double() - 0.5
        ret = th.Point3.exp_map(tangent_vector)

        assert torch.allclose(ret.tensor, tangent_vector, atol=EPS)
        check_projection_for_exp_map(
            tangent_vector, Group=th.Point3, is_projected=False
        )


def test_log_map():
    rng = torch.Generator()
    rng.manual_seed(0)

    for batch_size in [1, 20, 100]:
        group = th.Point2.rand(batch_size)
        ret = group.log_map()

        assert torch.allclose(ret, group.tensor, atol=EPS)
        check_projection_for_log_map(
            tangent_vector=ret, Group=th.Point2, is_projected=False
        )

    for batch_size in [1, 20, 100]:
        group = th.Point3.rand(batch_size)
        ret = group.log_map()

        assert torch.allclose(ret, group.tensor, atol=EPS)
        check_projection_for_log_map(
            tangent_vector=ret, Group=th.Point3, is_projected=False
        )


def test_local_map():
    rng = torch.Generator()
    rng.manual_seed(0)

    for batch_size in [1, 20, 100]:
        group0 = th.Point2.rand(batch_size)
        group1 = th.Point2.rand(batch_size)

        check_jacobian_for_local(group0, group1, Group=th.Point2, is_projected=False)

    for batch_size in [1, 20, 100]:
        group0 = th.Point3.rand(batch_size)
        group1 = th.Point3.rand(batch_size)

        check_jacobian_for_local(group0, group1, Group=th.Point3, is_projected=False)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import abc
from typing import List, Optional

from theseus.core import Objective

from .variable_ordering import VariableOrdering


class Linearization(abc.ABC):
    # if ordering is None, this will generate a default ordering
    def __init__(
        self,
        objective: Objective,
        ordering: Optional[VariableOrdering] = None,
        **kwargs
    ):
        self.objective = objective
        if ordering is None:
            ordering = VariableOrdering(objective, default_order=True)
        self.ordering = ordering
        if not self.ordering.complete:
            raise ValueError("Given variable ordering is not complete.")

        self.var_dims: List[int] = []
        self.var_start_cols: List[int] = []
        col_counter = 0
        for var in ordering:
            v_dim = var.dof()
            self.var_start_cols.append(col_counter)
            self.var_dims.append(v_dim)
            col_counter += v_dim

        self.num_cols = col_counter
        self.num_rows = self.objective.dim()

    @abc.abstractmethod
    def _linearize_jacobian_impl(self):
        pass

    @abc.abstractmethod
    def _linearize_hessian_impl(self):
        pass

    def linearize(self):
        if not self.ordering.complete:
            raise RuntimeError(
                "Attempted to linearize an objective with an incomplete variable order."
            )
        self._linearize_hessian_impl()

    def hessian_approx(self):
        raise NotImplementedError
